import inspect
from typing import dataclass_transform

from sonolus.backend.blocks import Block, PlayBlock, PreviewBlock, TutorialBlock, WatchBlock
from sonolus.backend.mode import Mode
from sonolus.script.internal.descriptor import SonolusDescriptor
from sonolus.script.internal.generic import validate_concrete_type
from sonolus.script.internal.simulation_context import sim_ctx
from sonolus.script.internal.value import Value


class _GlobalInfo:
    def __init__(self, name: str, size: int, blocks: dict[Mode, Block], offset: int | None):
        self.name = name
        self.size = size
        self.blocks = blocks
        self.offset = offset


class _GlobalField(SonolusDescriptor):
    def __init__(self, name: str, type_: type[Value], index: int, offset: int):
        self.name = name
        self.type = type_
        self.index = index
        self.offset = offset

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if sim_ctx():
            from sonolus.script.values import zeros

            return sim_ctx().get_or_put_value((instance, self), lambda: zeros(self.type))

        from sonolus.script.internal.context import ctx

        info = owner._global_info_
        if not ctx():
            raise RuntimeError("Global field access outside of compilation")
        base = ctx().get_global_base(info)
        return self.type._from_place_(base.add_offset(self.offset))._get_readonly_()

    def __set__(self, instance, value):
        from sonolus.script.internal.context import ctx

        if sim_ctx():
            from sonolus.script.values import zeros

            sim_ctx().set_or_put_value((instance, self), lambda: zeros(self.type), value)
            return

        info = instance._global_info_
        if not ctx():
            raise RuntimeError("Global field access outside of compilation")
        base = ctx().get_global_base(info)
        target = self.type._from_place_(base.add_offset(self.offset))
        if self.type._is_value_type_():
            target._set_(value)
        else:
            target._copy_from_(value)


class _GlobalPlaceholder:
    def __init__(self, type_: type[Value], blocks: dict[Mode, Block], offset: int | None):
        self.type = type_
        self.blocks = blocks
        self.offset = offset
        self.size = type_._size_()

    def get(self):
        from sonolus.script.internal.context import ctx

        if not ctx():
            raise RuntimeError("Global access outside of compilation")
        base = ctx().get_global_base(self)
        return self.type._from_place_(base)

    def _get_sim_replacement_(self):
        from sonolus.script.values import zeros

        return sim_ctx().get_or_put_value(self, lambda: zeros(self.type))


def _create_global(cls: type, blocks: dict[Mode, Block], offset: int | None):
    if issubclass(cls, Value):
        cls = validate_concrete_type(cls)
        return _GlobalPlaceholder(cls, blocks, offset)
    if len(cls.__bases__) != 1:
        raise TypeError("Expected a class with no bases or a Value subclass")
    field_offset = 0
    for i, (
        name,
        annotation,
    ) in enumerate(inspect.get_annotations(cls, eval_str=True).items()):
        type_ = validate_concrete_type(annotation)
        setattr(cls, name, _GlobalField(name, type_, i, field_offset))
        field_offset += type_._size_()
    cls._global_info_ = _GlobalInfo(cls.__name__, field_offset, blocks, offset)  # type: ignore
    cls._is_comptime_value_ = True  # type: ignore
    return cls()


@dataclass_transform()
def _play_runtime_environment[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.PLAY: PlayBlock.RuntimeEnvironment}, 0)


@dataclass_transform()
def _watch_runtime_environment[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.WATCH: WatchBlock.RuntimeEnvironment}, 0)


@dataclass_transform()
def _tutorial_runtime_environment[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.TUTORIAL: TutorialBlock.RuntimeEnvironment}, 0)


@dataclass_transform()
def _preview_runtime_environment[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.PREVIEW: PreviewBlock.RuntimeEnvironment}, 0)


@dataclass_transform()
def _play_runtime_update[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.PLAY: PlayBlock.RuntimeUpdate}, 0)


@dataclass_transform()
def _watch_runtime_update[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.WATCH: WatchBlock.RuntimeUpdate}, 0)


@dataclass_transform()
def _preview_runtime_canvas[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.PREVIEW: PreviewBlock.RuntimeCanvas}, 0)


@dataclass_transform()
def _tutorial_runtime_update[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.TUTORIAL: TutorialBlock.RuntimeUpdate}, 0)


@dataclass_transform()
def _runtime_touch_array[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.PLAY: PlayBlock.RuntimeTouchArray}, 0)


@dataclass_transform()
def _runtime_skin_transform[T](cls: type[T]) -> T:
    return _create_global(
        cls,
        {
            Mode.PLAY: PlayBlock.RuntimeSkinTransform,
            Mode.WATCH: WatchBlock.RuntimeSkinTransform,
            Mode.PREVIEW: PreviewBlock.RuntimeSkinTransform,
            Mode.TUTORIAL: TutorialBlock.RuntimeSkinTransform,
        },
        0,
    )


@dataclass_transform()
def _runtime_particle_transform[T](cls: type[T]) -> T:
    return _create_global(
        cls,
        {
            Mode.PLAY: PlayBlock.RuntimeParticleTransform,
            Mode.WATCH: WatchBlock.RuntimeParticleTransform,
            Mode.TUTORIAL: TutorialBlock.RuntimeParticleTransform,
        },
        0,
    )


@dataclass_transform()
def _runtime_background[T](cls: type[T]) -> T:
    return _create_global(
        cls,
        {
            Mode.PLAY: PlayBlock.RuntimeBackground,
            Mode.WATCH: WatchBlock.RuntimeBackground,
            Mode.TUTORIAL: TutorialBlock.RuntimeBackground,
        },
        0,
    )


@dataclass_transform()
def _play_runtime_ui[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.PLAY: PlayBlock.RuntimeUI}, 0)


@dataclass_transform()
def _watch_runtime_ui[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.WATCH: WatchBlock.RuntimeUI}, 0)


@dataclass_transform()
def _tutorial_runtime_ui[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.TUTORIAL: TutorialBlock.RuntimeUI}, 0)


@dataclass_transform()
def _preview_runtime_ui[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.PREVIEW: PreviewBlock.RuntimeUI}, 0)


@dataclass_transform()
def _play_runtime_ui_configuration[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.PLAY: PlayBlock.RuntimeUIConfiguration}, 0)


@dataclass_transform()
def _watch_runtime_ui_configuration[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.WATCH: WatchBlock.RuntimeUIConfiguration}, 0)


@dataclass_transform()
def _tutorial_runtime_ui_configuration[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.TUTORIAL: TutorialBlock.RuntimeUIConfiguration}, 0)


@dataclass_transform()
def _preview_runtime_ui_configuration[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.PREVIEW: PreviewBlock.RuntimeUIConfiguration}, 0)


@dataclass_transform()
def _tutorial_instruction[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.TUTORIAL: TutorialBlock.TutorialInstruction}, 0)


@dataclass_transform()
def level_memory[T](cls: type[T]) -> T:
    """Define level memory.

    Level memory may be modified during gameplay in sequential callbacks
    ([`preprocess`][sonolus.script.archetype.PlayArchetype.preprocess],
    [`update_sequential`][sonolus.script.archetype.PlayArchetype.update_sequential],
    [`touch`][sonolus.script.archetype.PlayArchetype.touch]).
    Compared to level data, it allows modification during gameplay, but prevents some optimizations.

    Usage:
        ```python
        @level_memory
        class LevelMemory:
            variable: int

        variable = level_memory(Array[int, 10])

        def f():
            LevelMemory.variable = 1
            variable[0] = 2
        ```
    """
    return _create_global(
        cls,
        {
            Mode.PLAY: PlayBlock.LevelMemory,
            Mode.WATCH: WatchBlock.LevelMemory,
            Mode.TUTORIAL: TutorialBlock.TutorialMemory,
        },
        None,
    )


@dataclass_transform()
def level_data[T](cls: type[T]) -> T:
    """Define level data.

    Level data may only be modified during [`preprocess`][sonolus.script.archetype.PlayArchetype.preprocess].
    Compared to level memory, it enables some optimizations during gameplay, so it's recommended to use it
    if mutation is only needed during preprocessing.

    Usage:
        ```python
        @level_data
        class LevelData:
            variable: int

        variable = level_data(Array[int, 10])

        def f():
            LevelData.variable = 1
            variable[0] = 2
        ```
    """
    return _create_global(
        cls,
        {
            Mode.PLAY: PlayBlock.LevelData,
            Mode.WATCH: WatchBlock.LevelData,
            Mode.PREVIEW: PreviewBlock.PreviewData,
            Mode.TUTORIAL: TutorialBlock.TutorialData,
        },
        None,
    )


# level_option is handled by the options decorator
# level_bucket is handled by the bucket decorator


@dataclass_transform()
def _level_score[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.PLAY: PlayBlock.LevelScore, Mode.WATCH: WatchBlock.LevelScore}, 0)


@dataclass_transform()
def _level_life[T](cls: type[T]) -> T:
    return _create_global(cls, {Mode.PLAY: PlayBlock.LevelLife, Mode.WATCH: WatchBlock.LevelLife}, 0)


# engine_rom is handled by the compiler
# entity memory is handled by the archetype
# entity data is handled by the archetype
# entity shared memory is handled by the archetype
# entity info is handled by the archetype
# entity despawn is handled by the archetype
# entity input is handled by the archetype
# entity data array is handled by the archetype
# entity shared memory array is handled by the archetype
# entity info array is handled by the archetype
# archetype life is handled by the archetype
# temporary memory is handled by the compiler



from __future__ import annotations

import inspect
from abc import ABCMeta
from collections.abc import Callable, Iterable
from functools import wraps
from inspect import getmro
from typing import Any, ClassVar, Protocol, Self, TypeVar, dataclass_transform, get_origin

from sonolus.backend.place import BlockPlace
from sonolus.script.internal.context import ctx
from sonolus.script.internal.descriptor import SonolusDescriptor
from sonolus.script.internal.generic import (
    GenericValue,
    accept_and_infer_types,
    validate_and_resolve_type,
    validate_concrete_type,
    validate_type_spec_with_extras,
)
from sonolus.script.internal.meta_fn import meta_fn
from sonolus.script.internal.value import BackingSource, DataValue, Value
from sonolus.script.num import Num


# Protocol metaclass is a subclass of ABCMeta, but let's make ABCMeta explicit for clarity
class RecordMeta(type(Protocol), ABCMeta):
    @meta_fn
    def __pos__[T](cls: type[T]) -> T:
        """Create a zero-initialized record instance."""
        return cls._zero_()


@dataclass_transform(eq_default=True)
class Record(GenericValue, metaclass=RecordMeta):
    """Base class for user-defined data structures.

    Usage:
        A regular record:
        ```python
        class MyRecord(Record):
            field1: int
            field2: bool
        ```

        A generic record:
        ```python
        class MyGenericRecord[T, U](Record):
            field1: T
            field2: U
        ```

        Creating an instance:
        ```python
        record = MyRecord(field1=42, field2=True)
        record_2 = MyGenericRecord[int, int](field1=42, field2=100)
        record_3 = MyGenericRecord(field1=42, field2=100)  # Type arguments can be inferred
        record_4 = +MyRecord  # Create a zero-initialized record
        record_5 = +MyGenericRecord[int, int]
        ```

        Copying a record:
        ```python
        record_copy = +record
        ```
    """

    _value_: dict[str, Value]
    _fields_: ClassVar[list[_RecordField] | None] = None
    _constructor_signature_: ClassVar[inspect.Signature]
    _is_frozen_: bool = False

    @classmethod
    @meta_fn
    def frozen[**P](cls: Callable[P, Self], /, *args: P.args, **kwargs: P.kwargs) -> Self:
        """Create a frozen record instance.

        Numeric fields in the record will not be writable. However, nested values such as other records may remain
        mutable.

        Args:
            *args: Positional arguments for the record constructor.
            **kwargs: Keyword arguments for the record constructor.

        Returns:
            A frozen record instance.
        """
        # We override __new__ to allow changing to the parameterized version
        if cls._constructor_signature_ is None:
            raise TypeError(f"Cannot instantiate {cls.__name__}")
        bound = cls._constructor_signature_.bind(*args, **kwargs)
        bound.apply_defaults()
        values = {}
        type_vars = {}
        for field in cls._fields_:
            value = bound.arguments[field.name]
            value = accept_and_infer_types(field.type, value, type_vars)
            values[field.name] = value._get_readonly_()
        for type_param in cls.__type_params__:
            if type_param not in type_vars:
                raise TypeError(f"Type parameter {type_param} cannot be inferred and must be provided explicitly")
        type_args = tuple(type_vars[type_param] for type_param in cls.__type_params__)
        if cls._type_args_ is not None:
            parameterized = cls
        else:
            parameterized = cls[type_args]
        result: cls = object.__new__(parameterized)  # type: ignore
        result._value_ = values
        result._is_frozen_ = True
        return result

    @classmethod
    def _validate_type_args_(cls, args: tuple[Any, ...]) -> tuple[Any, ...]:
        if cls._fields_ is None:
            raise TypeError("Base Record class cannot have type arguments")
        return super()._validate_type_args_(args)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()
        is_parameterizing = cls._type_args_ is not None and all(
            getattr(parent, "_type_args_", None) is None for parent in getmro(cls)[1:]
        )
        if is_parameterizing:
            fields = []
            offset = 0
            for generic_field in cls._fields_:
                resolved_type = validate_and_resolve_type(generic_field.type, cls._type_vars_to_args_)
                resolved_type = validate_concrete_type(resolved_type)
                field = _RecordField(
                    generic_field.name, resolved_type, generic_field.index, offset, generic_field.final
                )
                fields.append(field)
                setattr(cls, field.name, field)
                offset += resolved_type._size_()
            cls._fields_ = fields
            return
        is_inheriting_from_existing_record_class = cls._fields_ is not None
        if is_inheriting_from_existing_record_class and not is_parameterizing:
            # The main reason this is disallowed is that subclasses wouldn't be substitutable for their parent classes
            # Assignment of a subclass instance to a variable of the parent class would either be disallowed or would
            # require object slicing. Either way, it could lead to confusion.
            # Dealing with generic supertypes is also tricky, so it isn't really worth the effort to support this.
            raise TypeError("Subclassing of a Record is not supported")

        hints = inspect.get_annotations(cls, eval_str=True)
        fields = []
        params = []
        index = 0
        offset = 0
        for name, hint in hints.items():
            try:
                if name not in cls.__annotations__:
                    continue
                if hint is ClassVar or get_origin(hint) is ClassVar:
                    continue
                if hasattr(cls, name):
                    raise TypeError("Default values are not supported for Record fields")
                type_info = validate_type_spec_with_extras(hint)
                type_ = type_info.value
                fields.append(_RecordField(name, type_, index, offset, type_info.final))
                if isinstance(type_, type) and issubclass(type_, Value) and type_._is_concrete_():
                    offset += type_._size_()
                setattr(cls, name, fields[-1])
                index += 1
                params.append(
                    inspect.Parameter(
                        name,
                        inspect.Parameter.POSITIONAL_OR_KEYWORD,
                        annotation=type_,
                    )
                )
            except Exception as e:
                raise TypeError(f"Error processing field '{name}' of Record {cls.__name__}: {e}") from e

        cls._parameterized_ = {}
        cls._fields_ = fields
        cls._constructor_signature_ = inspect.Signature(params)

        _add_inplace_ops(cls)

        cls.__match_args__ = tuple(field.name for field in fields)

        if len(getattr(cls, "__type_params__", ())) == 0:
            # Make the class behave as the parameterized version
            cls._type_args_ = ()
            cls._type_vars_to_args_ = {}
            cls._parameterized_[()] = cls

    def __new__(cls, *args, **kwargs):
        # We override __new__ to allow changing to the parameterized version
        if cls._constructor_signature_ is None:
            raise TypeError(f"Cannot instantiate {cls.__name__}")
        bound = cls._constructor_signature_.bind(*args, **kwargs)
        bound.apply_defaults()
        values = {}
        type_vars = {}
        for field in cls._fields_:
            value = bound.arguments[field.name]
            value = accept_and_infer_types(field.type, value, type_vars)
            if field.final:
                values[field.name] = value._get_readonly_()
            else:
                values[field.name] = value._get_()
        for type_param in cls.__type_params__:
            if type_param not in type_vars:
                raise TypeError(f"Type parameter {type_param} cannot be inferred and must be provided explicitly")
        type_args = tuple(type_vars[type_param] for type_param in cls.__type_params__)
        if cls._type_args_ is not None:
            parameterized = cls
        else:
            parameterized = cls[type_args]
        result: cls = object.__new__(parameterized)  # type: ignore
        result._value_ = values
        return result

    def __init__(self, *args, **kwargs):
        # Initialization is done in __new__ and other methods
        pass

    @classmethod
    def _raw(cls, **kwargs) -> Self:
        result = object.__new__(cls)
        result._value_ = kwargs
        return result

    @classmethod
    def _unchecked(cls, **kwargs) -> Self:
        # Skips most validation, generally for internal use in frequently-called methods for performance reasons
        result = object.__new__(cls)
        for k, v in kwargs.items():
            if isinstance(v, int | float):
                kwargs[k] = Num._accept_(v)
        result._value_ = kwargs
        return result

    @classmethod
    def _size_(cls) -> int:
        return sum(field.type._size_() for field in cls._fields_)

    @classmethod
    def _is_value_type_(cls) -> bool:
        return False

    @classmethod
    def _from_backing_source_(cls, source: BackingSource) -> Self:
        result = object.__new__(cls)
        result._value_ = {
            field.name: field.type._from_backing_source_(
                lambda offset, field_offset=field.offset: source((Num(offset) + Num(field_offset)).ir())  # type: ignore
            )
            for field in cls._fields_
        }
        return result

    @classmethod
    def _from_place_(cls, place: BlockPlace) -> Self:
        result = object.__new__(cls)
        result._value_ = {field.name: field.type._from_place_(place.add_offset(field.offset)) for field in cls._fields_}
        return result

    @classmethod
    def _accepts_(cls, value: Any) -> bool:
        return issubclass(type(value), cls)

    @classmethod
    def _accept_(cls, value: Any) -> Self:
        if not cls._accepts_(value):
            raise TypeError(f"Cannot accept value {value} as {cls.__name__}")
        return value

    def _is_py_(self) -> bool:
        return all(value._is_py_() for value in self._value_.values())

    def _as_py_(self) -> Self:
        if not self._is_py_():
            raise ValueError("Not a python value")
        return self

    @classmethod
    def _from_list_(cls, values: Iterable[DataValue]) -> Self:
        iterator = iter(values)
        return cls._raw(**{field.name: field.type._from_list_(iterator) for field in cls._fields_})

    def _to_list_(self, level_refs: dict[Any, str] | None = None) -> list[DataValue | str]:
        result = []
        for field in self._fields_:
            result.extend(self._value_[field.name]._to_list_(level_refs))
        return result

    @classmethod
    def _flat_keys_(cls, prefix: str) -> list[str]:
        result = []
        for field in cls._fields_:
            result.extend(field.type._flat_keys_(f"{prefix}.{field.name}"))
        if len(result) == 1:
            return [prefix]
        return result

    def _get_(self) -> Self:
        return self

    def _set_(self, value: Any):
        raise TypeError("Record does not support _set_")

    def _copy_from_(self, value: Any):
        value = self._accept_(value)
        for field in self._fields_:
            field.__set__(self, field.__get__(value))

    def _copy_(self) -> Self:
        return type(self)._raw(**{field.name: self._value_[field.name]._copy_() for field in self._fields_})

    @classmethod
    def _alloc_(cls) -> Self:
        # Compared to using the constructor, this avoids unnecessary _get_ calls
        result = object.__new__(cls)
        result._value_ = {field.name: field.type._alloc_() for field in cls._fields_}
        return result

    @classmethod
    def _zero_(cls) -> Self:
        result = object.__new__(cls)
        result._value_ = {field.name: field.type._zero_() for field in cls._fields_}
        return result

    def __str__(self):
        return f"{self.__class__.__name__}({
            ', '.join(f'{field.name}={field.get_internal(self)}' for field in self._fields_)
        })"

    def __repr__(self):
        return f"{self.__class__.__name__}({
            ', '.join(f'{field.name}={field.get_internal(self)!r}' for field in self._fields_)
        })"

    @meta_fn
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False
        result: Num = Num._accept_(True)
        for field in self._fields_:
            result = result.and_(field.__get__(self) == field.__get__(other))
        return result

    @meta_fn
    def __ne__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return True
        result: Num = Num._accept_(False)
        for field in self._fields_:
            result = result.or_(field.__get__(self) != field.__get__(other))
        return result

    def __hash__(self):
        return hash(tuple(field.__get__(self) for field in self._fields_))

    @meta_fn
    def __pos__(self) -> Self:
        """Return a copy of the record."""
        return self._copy_()

    @classmethod
    @meta_fn
    def type_var_value(cls, var: TypeVar, /) -> Any:
        """Return the value of a type variable.

        Args:
            var: The type variable to get the value of.

        Returns:
            The value of the type variable.
        """
        return super().type_var_value(var)

    @classmethod
    def _validate_parameterized_(cls):
        pass

    @classmethod
    def _get_parameterized(cls, args: tuple[Any, ...]) -> type[Self]:
        result = super()._get_parameterized(args)
        result._validate_parameterized_()
        return result


class _RecordField(SonolusDescriptor):
    def __init__(self, name: str, type_: type[Value] | Any, index: int, offset: int, final: bool):
        self.name = name
        self.type = type_
        self.index = index
        self.offset = offset
        self.final = final

    def get_internal(self, instance: Record) -> Value:
        return instance._value_[self.name]

    def __get__(self, instance: Record | None, owner=None):
        if instance is None:
            return self
        if instance._is_frozen_ or self.final:
            result = instance._value_[self.name]
        else:
            result = instance._value_[self.name]._get_readonly_()
        if ctx():
            return result
        else:
            return result._as_py_()

    def __set__(self, instance: Record, value):
        if instance._is_frozen_:
            raise TypeError("Cannot set fields of a frozen Record")
        if self.final:
            raise TypeError("Cannot set a final field")
        value = self.type._accept_(value)
        if self.type._is_value_type_():
            instance._value_[self.name]._set_(value)
        else:
            instance._value_[self.name]._copy_from_(value)


_ops_to_inplace_ops = {
    "__add__": "__iadd__",
    "__sub__": "__isub__",
    "__mul__": "__imul__",
    "__truediv__": "__itruediv__",
    "__floordiv__": "__ifloordiv__",
    "__mod__": "__imod__",
    "__pow__": "__ipow__",
    "__lshift__": "__ilshift__",
    "__rshift__": "__irshift__",
    "__or__": "__ior__",
    "__xor__": "__ixor__",
    "__and__": "__iand__",
    "__matmul__": "__imatmul__",
}


def _add_inplace_ops(cls):
    for op, inplace_op in _ops_to_inplace_ops.items():
        # Note that hasattr doesn't work, since it'll return True for metaclass methods
        if op in cls.__dict__ and not hasattr(cls, inplace_op):
            setattr(cls, inplace_op, _make_inplace_op(op, getattr(cls, op)))
    return cls


def _make_inplace_op(op: str, orig_fn):
    @meta_fn
    @wraps(orig_fn)
    def inplace_op(self, other):
        _compiler_internal_ = True  # noqa: F841
        self._copy_from_(getattr(self, op)(other))
        return self

    return inplace_op



# ruff: noqa
import builtins
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Sequence,
    overload,
)

def all(iterable: Iterable[builtins.bool]) -> builtins.bool:
    """Return True if all elements of the iterable are true.

    Args:
        iterable: The iterable to evaluate.

    Returns:
        True if all elements are true, False otherwise.
    """
    ...

def any(iterable: Iterable[builtins.bool]) -> builtins.bool:
    """Return True if any element of the iterable is true.

    Args:
        iterable: The iterable to evaluate.

    Returns:
        True if any element is true, False otherwise.
    """
    ...

def sum(
    iterable: Iterable[builtins.int | builtins.float], /, start: builtins.int | builtins.float = 0
) -> builtins.int | builtins.float:
    """Return the sum of a 'start' value (default: 0) and an iterable of numbers.

    Args:
        iterable: The iterable of numbers to sum.
        start: The starting value to add to the sum.

    Returns:
        The total sum.
    """
    ...

def abs(x: builtins.int | builtins.float) -> builtins.int | builtins.float:
    """Return the absolute value of a number.

    Args:
        x: A number.

    Returns:
        The absolute value of x.
    """
    ...

def bool(x: builtins.int | builtins.float | builtins.bool) -> builtins.bool:
    """Convert a value to a Boolean.

    Args:
        x: The value to convert.

    Returns:
        The Boolean value of x.
    """
    ...

def callable(obj: object) -> bool:
    """Check if the object appears callable.

    Args:
        obj: The object to check.

    Returns:
        True if the object appears callable, False otherwise.
    """
    ...

@overload
def dict() -> builtins.dict: ...
@overload
def dict[K, V](mapping: builtins.dict[K, V], **kwargs: V) -> builtins.dict[K, V]: ...
@overload
def dict[K, V](iterable: Iterable[tuple[K, V]], **kwargs: V) -> builtins.dict[K, V]: ...
@overload
def dict[V](**kwargs: V) -> builtins.dict[builtins.str, V]: ...
def dict(*args, **kwargs) -> builtins.dict:
    """Construct a dict from a mapping, an iterable of key-value pairs, or keyword arguments.

    All dict keys must be compile-time constants. Dynamic access using a key that is not
    a compile-time constant is only supported for numeric, Array, and Record values.

    Args:
        mapping_or_iterable: An optional dict to copy from, or an iterable of
            ``(key, value)`` pairs.
        **kwargs: Optional keyword arguments to include in the dict.

    Returns:
        A new dict.
    """
    ...

def enumerate[T](iterable: Iterable[T], start: int = 0) -> Iterator[tuple[int, T]]:
    """Return an enumerate object.

    Args:
        iterable: The iterable to enumerate.
        start: The starting index.

    Returns:
        An enumerate object.
    """
    ...

def filter[T](function: Callable[[T], builtins.bool] | None, iterable: Iterable[T]) -> Iterator[T]:
    """Construct an iterator from those elements of iterable for which function returns true.

    Args:
        function: A function that tests if each element should be included. If None, returns the elements that are true.
        iterable: The iterable to filter.

    Returns:
        An iterator yielding the filtered elements.
    """
    ...

def float(x: builtins.int | builtins.float) -> builtins.float:
    """Convert a number to a floating point number.

    Args:
        x: The number to convert.

    Returns:
        The floating point representation of x.
    """
    ...

@overload
def getattr(obj: object, name: builtins.str) -> Any: ...
@overload
def getattr[T](obj: object, name: builtins.str, default: T) -> Any | T: ...
def getattr(obj: object, name: builtins.str, default: Any = ...) -> Any:
    """Get a named attribute from an object.

    Args:
        obj: The object to get the attribute from.
        name: The name of the attribute.
        default: The value to return if the attribute does not exist.

    Returns:
        The value of the attribute, or default if it does not exist.
    """
    ...

def hasattr(obj: object, name: builtins.str) -> builtins.bool:
    """Check if an object has a named attribute.

    Can result in a compile-time error if an unsupported property is checked,
    since later trying to access it would also fail.

    Args:
        obj: The object to check.
        name: The name of the attribute.

    Returns:
        True if the object has the attribute, False otherwise.
    """
    ...

def int(x: builtins.int | builtins.float) -> builtins.int:
    """Convert a number to an integer.

    Args:
        x: The number to convert.

    Returns:
        The integer representation of x.
    """
    ...

def isinstance(obj: object, classinfo: type | tuple[type, ...]) -> builtins.bool:
    """Check if an object is an instance of a class or of a subclass thereof.

    Args:
        obj: The object to check.
        classinfo: A type or a tuple of types.

    Returns:
        True if the object is an instance of classinfo, False otherwise.
    """
    ...

def issubclass(cls: type, classinfo: type | tuple[type, ...]) -> builtins.bool:
    """Check if a class is a subclass of another class or a tuple of classes.

    Args:
        cls: The class to check.
        classinfo: A class or a tuple of classes.

    Returns:
        True if cls is a subclass of classinfo, False otherwise.
    """
    ...

def iter[T](iterable: Iterable[T]) -> Iterator[T]:
    """Return an iterator for the given iterable.

    Args:
        iterable: The iterable to convert to an iterator.

    Returns:
        An iterator over the elements of the iterable.
    """
    ...

def len(s: object) -> builtins.int:
    """Return the number of items in a container.

    Args:
        s: The container object.

    Returns:
        The number of items in s.
    """
    ...

def map[T, S](function: Callable[[T], S], iterable: Iterable[T]) -> Iterator[S]:
    """Apply a function to every item of an iterable and return an iterator.

    Args:
        function: The function to apply.
        iterable: The iterable to process.

    Returns:
        An iterator with the results.
    """
    ...

@overload
def max[T](iterable: Iterable[T], *, key: Callable[[T], Any] | None = ...) -> T: ...
@overload
def max(
    iterable: Iterable[builtins.int | builtins.float],
    *,
    default: builtins.int | builtins.float = ...,
    key: Callable[[builtins.int | builtins.float], Any] | None = ...,
) -> builtins.int | builtins.float: ...
@overload
def max[T](arg1: T, arg2: T, *args: T, key: Callable[[T], Any] | None = ...) -> T: ...
def max(*args, **kwargs):
    """Return the largest item in an iterable or the largest of multiple arguments.

    When called with a single iterable, returns the largest item from that iterable.
    When called with multiple arguments, returns the largest argument.

    Use the `key` parameter to specify a function that transforms each element
    before comparison. Use the `default` parameter to specify a value to return
    if the iterable is empty (supported only for numeric values).
    """
    ...

@overload
def min[T](iterable: Iterable[T], *, key: Callable[[T], Any] | None = ...) -> T: ...
@overload
def min(
    iterable: Iterable[builtins.int | builtins.float],
    *,
    default: builtins.int | builtins.float = ...,
    key: Callable[[builtins.int | builtins.float], Any] | None = ...,
) -> builtins.int | builtins.float: ...
@overload
def min[T](arg1: T, arg2: T, *args: T, key: Callable[[T], Any] | None = ...) -> T: ...
def min(*args, **kwargs):
    """Return the smallest item in an iterable or the smallest of multiple arguments.

    When called with a single iterable, returns the smallest item from that iterable.
    When called with multiple arguments, returns the smallest argument.

    Use the `key` parameter to specify a function that transforms each element
    before comparison. Use the `default` parameter to specify a value to return
    if the iterable is empty (supported only for numeric values).
    """
    ...

def next[T](iterator: Iterator[T]) -> T:
    """Retrieve the next item from an iterator.

    Errors if the iterator is exhausted.

    Args:
        iterator: The iterator to retrieve the next item from.

    Returns:
        The next item from the iterator.
    """
    ...

@overload
def range(stop: builtins.int) -> builtins.range: ...
@overload
def range(start: builtins.int, stop: builtins.int, step: builtins.int = ...) -> builtins.range: ...
def range(*args) -> builtins.range:
    """Return an immutable sequence of numbers.

    When called with one argument, creates a sequence from 0 to that number (exclusive).
    When called with two arguments, creates a sequence from the first to the second (exclusive).
    When called with three arguments, the third argument specifies the step size.
    """
    ...

def reversed[T](seq: Sequence[T]) -> Iterator[T]:
    """Return a reverse iterator.

    Args:
        seq: The sequence to reverse.

    Returns:
        An iterator over the reversed sequence.
    """
    ...

def round(number: builtins.int | builtins.float, ndigits: builtins.int = ...) -> builtins.float:
    """Round a number to a given precision in decimal digits.

    Args:
        number: The number to round.
        ndigits: The number of decimal digits to round to.

    Returns:
        The rounded number.
    """
    ...

@overload
def set() -> builtins.set: ...
@overload
def set[T](iterable: Iterable[T]) -> builtins.set[T]: ...
def set(*args) -> builtins.set:
    """Construct a set from an iterable.

    All set members must be compile-time constants.

    Args:
        iterable: An optional iterable of compile-time constant values.

    Returns:
        A new set.
    """
    ...

def setattr(obj: object, name: builtins.str, value: Any) -> None:
    """Set a named attribute on an object.

    Args:
        obj: The object to set the attribute on.
        name: The name of the attribute.
        value: The value to set.

    Returns:
        None.
    """
    ...

def super(cls: type = ..., instance: Any = ...) -> Any:
    """Return a proxy object that delegates method calls to a parent or sibling class.

    Args:
        cls: The class to delegate.
        instance: The instance to delegate to.
    Returns:
        A proxy object that can be used to call methods from the parent or sibling class.
    """
    ...

def type(obj: object) -> builtins.type:
    """Return the type of an object.

    Args:
        obj: The object to get the type of.

    Returns:
        The type of the object.
    """
    ...

def zip[T](*iterables: Iterable[T]) -> Iterator[tuple[T, ...]]:
    """Return an iterator of tuples, where the i-th tuple contains the i-th element from each of the argument sequences.

    Args:
        *iterables: Iterables to aggregate.

    Returns:
        An iterator of aggregated tuples.
    """
    ...



import operator
from contextlib import contextmanager
from functools import reduce

import numpy as np
from hypothesis import example, given
from hypothesis import strategies as st

from sonolus.script import numtools
from sonolus.script.containers import Pair
from sonolus.script.num import _is_num  # noqa: PLC2701
from sonolus.script.numtools import (
    _ints_to_uint32,  # noqa: PLC2701
    _UInt32,  # noqa: PLC2701
    _uint32_to_comparable_float,  # noqa: PLC2701
    make_comparable_float,
    product,
    quantize_to_step,
)
from tests.script.conftest import run_and_validate

numtools.enable_np = True


class PlainField:
    def __init__(self, name):
        self.name = name

    def __get__(self, instance, owner=None):
        result = instance._value_[self.name]
        if _is_num(result):
            return np.float32(result._as_py_())
        return result

    def __set__(self, instance, value):
        if _is_num(value):
            value = np.float32(value._as_py_())
        instance._value_[self.name] = value


@contextmanager
def patch_float32_records(*classes):
    classes_to_orig_fields = {}
    for cls in classes:
        classes_to_orig_fields[cls] = cls._fields_
        new_fields = [PlainField(field.name) for field in cls._fields_]
        cls._fields_ = new_fields
        for field in new_fields:
            setattr(cls, field.name, field)

        def __new__(cls, *args, **kwargs):  # noqa: N807
            bound = cls._constructor_signature_.bind(*args, **kwargs)
            result = object.__new__(cls)
            result._value_ = {name: np.float32(value) for name, value in bound.arguments.items()}
            return result

        def _copy_(self):
            new_instance = object.__new__(self.__class__)
            new_instance._value_ = self._value_
            return new_instance

        cls.__new__ = __new__
        cls._copy_ = _copy_
    try:
        yield
    finally:
        for cls in classes:
            cls._fields_ = classes_to_orig_fields[cls]
            for field in cls._fields_:
                setattr(cls, field.name, field)
            del cls.__new__
            del cls._copy_


def patch():
    return patch_float32_records(_UInt32)


def uint32_to_int(u: _UInt32) -> int:
    return int(u.hi) << 16 | int(u.lo)


def int_to_uint32(value: int) -> _UInt32:
    hi = (value >> 16) & 0xFFFF
    lo = value & 0xFFFF
    return _UInt32(hi=hi, lo=lo)


@given(x=st.integers(min_value=0, max_value=2**31 - 2))
@example(x=0)
@example(x=2**30 - 1)
@example(x=2**30)
@example(x=2**30 + 1)
@example(x=2**31 - 2)
def test_uint32_to_comparable_float_with_float32(x: int):
    with patch():
        ux = int_to_uint32(x)
        uy = int_to_uint32(x + 1)
        fx = _uint32_to_comparable_float(ux)
        fy = _uint32_to_comparable_float(uy)
        assert fx < fy


@given(x=st.integers(min_value=0, max_value=2**31 - 2))
@example(x=0)
@example(x=2**30 - 1)
@example(x=2**30)
@example(x=2**30 + 1)
@example(x=2**31 - 2)
def test_uint32_to_comparable_float(x: int):
    ux = int_to_uint32(x)
    uy = int_to_uint32(x + 1)

    def fn():
        fx = _uint32_to_comparable_float(ux)
        fy = _uint32_to_comparable_float(uy)
        return fx < fy

    assert run_and_validate(fn)


@st.composite
def make_ints_to_uint32_arg(draw, max_max_value) -> tuple[int, int]:
    max_value = draw(st.integers(min_value=1, max_value=max_max_value))
    value = draw(st.integers(min_value=0, max_value=max_value - 1))
    return value, max_value


@st.composite
def make_ints_to_uint32_args(draw, min_args: int = 0, max_args: int = 5) -> tuple[tuple[int, int], ...]:
    headroom = (1 << 31) - 1
    args = []
    count = draw(st.integers(min_value=min_args, max_value=max_args))
    for _ in range(count):
        value, max_value = draw(make_ints_to_uint32_arg(max_max_value=headroom))
        args.append((value, max_value))
        headroom //= max_value
    return tuple(args)


@st.composite
def make_paired_ints_to_uint32_args(
    draw, min_args: int = 1, max_args: int = 5
) -> tuple[tuple[tuple[int, int], ...], tuple[tuple[int, int], ...]]:
    num_args = draw(st.integers(min_value=min_args, max_value=max_args))
    return (
        draw(make_ints_to_uint32_args(min_args=num_args, max_args=num_args)),
        draw(make_ints_to_uint32_args(min_args=num_args, max_args=num_args)),
    )


@given(args=make_ints_to_uint32_args())
def test_ints_to_uint32(args):
    def fn():
        return _ints_to_uint32(*args)

    expected = 0
    multiplier = 1
    for value, max_value in reversed(args):
        expected += value * multiplier
        multiplier *= max_value
    assert uint32_to_int(run_and_validate(fn)) == expected


@given(args=make_ints_to_uint32_args())
def test_ints_to_uint32_with_float32(args):
    with patch():
        expected = 0
        multiplier = 1
        for value, max_value in reversed(args):
            expected += value * multiplier
            multiplier *= max_value

        assert uint32_to_int(_ints_to_uint32(*args)) == expected


@given(args=make_paired_ints_to_uint32_args())
def test_ints_to_uint32_comparison(args: tuple[tuple[tuple[int, int], ...], tuple[tuple[int, int], ...]]):
    args1, args2 = args

    def fn():
        return Pair(_ints_to_uint32(*args1), _ints_to_uint32(*args2))

    u1, u2 = run_and_validate(fn).tuple
    r1 = uint32_to_int(u1)
    r2 = uint32_to_int(u2)
    assert (r1 < r2) == (u1 < u2)
    assert (r1 <= r2) == (u1 <= u2)
    assert (r1 > r2) == (u1 > u2)
    assert (r1 >= r2) == (u1 >= u2)
    assert (r1 == r2) == (u1 == u2)
    assert (r1 != r2) == (u1 != u2)


@given(args=make_paired_ints_to_uint32_args())
def test_ints_to_uint32_comparison_with_float32(args: tuple[tuple[tuple[int, int], ...], tuple[tuple[int, int], ...]]):
    args1, args2 = args
    with patch():
        u1 = _ints_to_uint32(*args1)
        u2 = _ints_to_uint32(*args2)
        r1 = uint32_to_int(u1)
        r2 = uint32_to_int(u2)
        assert (r1 < r2) == (u1 < u2)
        assert (r1 <= r2) == (u1 <= u2)
        assert (r1 > r2) == (u1 > u2)
        assert (r1 >= r2) == (u1 >= u2)
        assert (r1 == r2) == (u1 == u2)
        assert (r1 != r2) == (u1 != u2)


@given(args=make_paired_ints_to_uint32_args())
def test_make_comparable_float_comparison(args: tuple[tuple[tuple[int, int], ...], tuple[tuple[int, int], ...]]):
    args1, args2 = args

    def fn():
        return Pair(make_comparable_float(*args1), make_comparable_float(*args2))

    f1, f2 = run_and_validate(fn).tuple
    u1 = _ints_to_uint32(*args1)
    u2 = _ints_to_uint32(*args2)
    assert (f1 < f2) == (u1 < u2)
    assert (f1 <= f2) == (u1 <= u2)
    assert (f1 > f2) == (u1 > u2)
    assert (f1 >= f2) == (u1 >= u2)
    assert (f1 == f2) == (u1 == u2)
    assert (f1 != f2) == (u1 != u2)


@given(args=make_paired_ints_to_uint32_args())
def test_make_comparable_float_comparison_with_float32(
    args: tuple[tuple[tuple[int, int], ...], tuple[tuple[int, int], ...]],
):
    args1, args2 = args
    with patch():
        f1 = make_comparable_float(*args1)
        f2 = make_comparable_float(*args2)
        u1 = _ints_to_uint32(*args1)
        u2 = _ints_to_uint32(*args2)
        assert (f1 < f2) == (u1 < u2)
        assert (f1 <= f2) == (u1 <= u2)
        assert (f1 > f2) == (u1 > u2)
        assert (f1 >= f2) == (u1 >= u2)
        assert (f1 == f2) == (u1 == u2)
        assert (f1 != f2) == (u1 != u2)


@st.composite
def make_quantize_to_step_bounds(draw) -> tuple[float, float, float]:
    start = draw(st.floats(allow_infinity=False, allow_nan=False, min_value=-1e5, max_value=1e5))
    delta = draw(st.floats(allow_infinity=False, allow_nan=False, min_value=1e-5, max_value=1e5))
    step = draw(st.floats(allow_infinity=False, allow_nan=False, min_value=1e-2, max_value=1e5))
    stop = start + delta
    return start, stop, step


@st.composite
def make_in_range_quantize_to_step_args(draw) -> tuple[float, float, float, float]:
    start, stop, step = draw(make_quantize_to_step_bounds())
    value = draw(
        st.floats(allow_infinity=False, allow_nan=False, min_value=start, max_value=max(start, stop - step / 2))
    )
    return value, start, stop, step


@st.composite
def make_out_of_range_quantize_to_step_args(draw) -> tuple[float, float, float, float]:
    start, stop, step = draw(make_quantize_to_step_bounds())
    if draw(st.booleans()):
        value = draw(st.floats(allow_infinity=False, allow_nan=False, max_value=start - 1, min_value=-1e10))
    else:
        value = draw(st.floats(allow_infinity=False, allow_nan=False, min_value=stop + 1, max_value=1e10))
    return value, start, stop, step


@given(args=make_in_range_quantize_to_step_args())
def test_quantize_to_step_in_range(args: tuple[float, float, float, float]):
    value, start, stop, step = args

    def fn():
        return Pair(*quantize_to_step(value, start, stop, step))

    result_step, result_max_steps = run_and_validate(fn).tuple
    assert abs((start + result_step * step) - value) < step / 2 + 1e-4
    assert result_max_steps >= 1
    assert start + (result_max_steps - 1) * step < stop + 1e-4
    assert start + result_max_steps * step >= stop - 1e-4


@given(args=make_out_of_range_quantize_to_step_args())
def test_quantize_to_step_out_of_range(args: tuple[float, float, float, float]):
    value, start, stop, step = args

    def fn():
        return Pair(*quantize_to_step(value, start, stop, step))

    result_step, result_max_steps = run_and_validate(fn).tuple
    if value < start:
        assert result_step == 0
    else:
        assert result_step == result_max_steps - 1
    assert result_max_steps >= 1
    assert start + (result_max_steps - 1) * step < stop + 1e-4
    assert start + result_max_steps * step >= stop - 1e-4


@given(
    values=st.lists(
        st.floats(allow_infinity=False, allow_nan=False, min_value=1e-5, max_value=1e5), min_size=0, max_size=10
    )
)
def test_product(values: list[float]):
    values = tuple(values)

    def fn():
        return product(values)

    expected = reduce(operator.mul, values, 1.0)
    assert run_and_validate(fn) == expected



from pathlib import Path

DATA_DIR = Path(__file__).parent / "data"


def compare_with_reference(name: str, data: str):
    existing_path = DATA_DIR / f"{name}"
    if existing_path.exists():
        assert data == existing_path.read_bytes().decode("utf-8")
    else:
        existing_path.parent.mkdir(parents=True, exist_ok=True)
        existing_path.write_bytes(data.encode("utf-8"))



from abc import abstractmethod


class SonolusDescriptor:
    """Base class for Sonolus descriptors.

    The compiler checks if a descriptor is an instance of a subclass of this class,
    so it knows that it's a supported descriptor.
    """

    @abstractmethod
    def __get__(self, instance, owner):
        pass

    @abstractmethod
    def __set__(self, instance, value):
        pass



from __future__ import annotations

from enum import IntEnum

from pydori.lib.buckets import Buckets
from pydori.lib.effect import Effects
from pydori.lib.layer import LAYER_ARROW, LAYER_NOTE, LAYER_NOTE_HEAD, get_z
from pydori.lib.layout import (
    Layout,
    layout_directional_flick_arrow,
    layout_flick_arrow,
    layout_note_body,
    layout_note_circular_particle,
    layout_note_linear_particle,
    note_y_to_alpha,
)
from pydori.lib.options import Options
from pydori.lib.particle import Particles
from pydori.lib.skin import Skin
from pydori.lib.stage import play_lane_particle
from sonolus.script.archetype import PlayArchetype, WatchArchetype
from sonolus.script.bucket import Bucket, Judgment
from sonolus.script.easing import ease_out_quad
from sonolus.script.effect import Effect, LoopedEffectHandle
from sonolus.script.interval import interp_clamped
from sonolus.script.particle import Particle, ParticleHandle
from sonolus.script.runtime import time
from sonolus.script.sprite import Sprite

# Time spent by the flick arrow fading in each animation cycle.
FLICK_FADE_IN_TIME = 0.1

# Time spent by the flick arrow fading out each animation cycle.
FLICK_FADE_OUT_TIME = 0.1

# Length of one cycle of the flick arrow animation.
FLICK_ARROW_PERIOD = 0.3

# Length of one cycle of the directional flick arrow animation.
DIRECTIONAL_FLICK_ARROW_PERIOD = 0.3

# Duration of note particles.
NOTE_PARTICLE_DURATION = 0.5

# Duration of active_hold_note particles.
ACTIVE_HOLD_PARTICLE_DURATION = 1.5

# Minimum speed for a flick input to be counted in terms of lane widths per second.
FLICK_SPEED_THRESHOLD = 6.0

# Minimum speed for a directional flick input to be counted in terms of lane widths per second.
# The total threshold is calculated as base + increment * abs(direction).
DIRECTIONAL_FLICK_SPEED_THRESHOLD_BASE = 2.0
DIRECTIONAL_FLICK_SPEED_THRESHOLD_INCREMENT = 2.0


class NoteKind(IntEnum):
    TAP = 1
    FLICK = 2
    DIRECTIONAL_FLICK = 3
    HOLD_HEAD = 4
    HOLD_TICK = 5
    HOLD_ANCHOR = 6  # Invisible 'ticks' for curved hold notes
    HOLD_END = 7


def get_note_body_sprite(kind: NoteKind, direction: int) -> Sprite:
    result = +Sprite
    match kind:
        case NoteKind.TAP:
            result @= Skin.tap_note
        case NoteKind.FLICK:
            result @= Skin.flick_note
        case NoteKind.DIRECTIONAL_FLICK:
            if direction > 0:
                result @= Skin.right_flick_note
            else:
                result @= Skin.left_flick_note
        case NoteKind.HOLD_HEAD:
            result @= Skin.hold_head_note
        case NoteKind.HOLD_TICK:
            result @= Skin.hold_tick_note
        case NoteKind.HOLD_ANCHOR:
            pass
        case NoteKind.HOLD_END:
            result @= Skin.hold_end_note
    return result


def get_note_arrow_sprite(kind: NoteKind, direction: int) -> Sprite:
    result = +Sprite
    match kind:
        case NoteKind.FLICK:
            result @= Skin.flick_arrow
        case NoteKind.DIRECTIONAL_FLICK:
            if direction > 0:
                result @= Skin.right_flick_arrow
            else:
                result @= Skin.left_flick_arrow
        case _:
            pass
    return result


def get_note_linear_particle(kind: NoteKind, direction: int) -> Particle:
    result = +Particle
    match kind:
        case NoteKind.TAP:
            result @= Particles.tap_linear
        case NoteKind.FLICK:
            result @= Particles.flick_linear
        case NoteKind.DIRECTIONAL_FLICK:
            if direction > 0:
                result @= Particles.right_flick_linear
            else:
                result @= Particles.left_flick_linear
        case NoteKind.HOLD_HEAD | NoteKind.HOLD_TICK | NoteKind.HOLD_END:
            result @= Particles.hold_linear
        case _:
            pass
    return result


def get_note_circular_particle(kind: NoteKind, direction: int) -> Particle:
    result = +Particle
    match kind:
        case NoteKind.TAP:
            result @= Particles.tap_circular
        case NoteKind.FLICK:
            result @= Particles.flick_circular
        case NoteKind.DIRECTIONAL_FLICK:
            if direction > 0:
                result @= Particles.right_flick_circular
            else:
                result @= Particles.left_flick_circular
        case NoteKind.HOLD_HEAD | NoteKind.HOLD_TICK | NoteKind.HOLD_END:
            result @= Particles.hold_circular
        case _:
            pass
    return result


def get_note_head_sprite() -> Sprite:
    """Return the sprite for hold note heads.

    Used by connectors to draw the head of an active hold note at the judgment line.
    """
    return Skin.hold_head_note


def get_note_active_circular_particle() -> Particle:
    """Return the active circular particle effect for hold notes.

    Used by connectors to draw the active circular particle effect for hold notes.
    """
    return Particles.hold_active_circular


def get_note_bucket(kind: NoteKind) -> Bucket:
    result = +Bucket
    match kind:
        case NoteKind.TAP:
            result @= Buckets.tap_note
        case NoteKind.FLICK:
            result @= Buckets.flick_note
        case NoteKind.DIRECTIONAL_FLICK:
            result @= Buckets.directional_flick_note
        case NoteKind.HOLD_HEAD:
            result @= Buckets.hold_head_note
        case NoteKind.HOLD_TICK:
            result @= Buckets.hold_tick_note
        case NoteKind.HOLD_ANCHOR:
            pass  # Hold anchors do not have a bucket
        case NoteKind.HOLD_END:
            result @= Buckets.hold_end_note
    return result


def draw_note(
    kind: NoteKind,
    lane: float,
    y: float,
    direction: int = 0,
):
    if kind == NoteKind.HOLD_ANCHOR:
        return
    body_sprite = get_note_body_sprite(kind, direction)
    arrow_sprite = get_note_arrow_sprite(kind, direction)
    draw_note_body(body_sprite, lane, y)
    if kind == NoteKind.FLICK:
        draw_flick_arrow(arrow_sprite, lane, y)
    elif kind == NoteKind.DIRECTIONAL_FLICK:
        draw_directional_flick_arrow(arrow_sprite, lane, y, direction)


def draw_note_body(sprite: Sprite, lane: float, y: float):
    alpha = note_y_to_alpha(y)
    if alpha <= 0:
        return
    layout = layout_note_body(lane, y)
    sprite.draw(layout, z=get_z(LAYER_NOTE, lane=lane, y=y), a=alpha)


def draw_flick_arrow(sprite: Sprite, lane: float, y: float):
    cycle_time = time() % FLICK_ARROW_PERIOD
    progress = cycle_time / FLICK_ARROW_PERIOD
    alpha = note_y_to_alpha(y) * ease_out_quad(
        interp_clamped(
            (
                0,
                FLICK_FADE_IN_TIME,
                FLICK_ARROW_PERIOD - FLICK_FADE_OUT_TIME,
                FLICK_ARROW_PERIOD,
            ),
            (
                0,
                1,
                1,
                0,
            ),
            cycle_time,
        )
    )
    if alpha <= 0:
        return
    layout = layout_flick_arrow(lane, y, progress)
    sprite.draw(layout, z=get_z(LAYER_ARROW, lane=lane, y=y), a=alpha)


def draw_directional_flick_arrow(sprite: Sprite, lane: float, y: float, direction: int):
    cycle_time = time() % DIRECTIONAL_FLICK_ARROW_PERIOD
    progress = cycle_time / DIRECTIONAL_FLICK_ARROW_PERIOD
    base_alpha = note_y_to_alpha(y)
    for i in range(abs(direction)):
        alpha = base_alpha * ease_out_quad(
            interp_clamped(
                (
                    0,
                    FLICK_FADE_IN_TIME,
                    DIRECTIONAL_FLICK_ARROW_PERIOD - FLICK_FADE_OUT_TIME,
                    DIRECTIONAL_FLICK_ARROW_PERIOD,
                ),
                (
                    0 if i == 0 else 1,  # Fade in only for the first arrow
                    1,
                    1,
                    0 if i == abs(direction) - 1 else 1,  # Fade out only for the last arrow
                ),
                cycle_time,
            )
        )
        if alpha <= 0:
            continue
        layout = layout_directional_flick_arrow(lane, y, direction, i, progress)
        sprite.draw(layout, z=get_z(LAYER_ARROW, lane=lane, y=y), a=alpha)


def play_note_particle(
    kind: NoteKind,
    lane: float,
    direction: int = 0,
):
    if kind == NoteKind.HOLD_ANCHOR:
        return
    if not Options.note_effect_enabled:
        return
    linear_particle = get_note_linear_particle(kind, direction)
    linear_layout = layout_note_linear_particle(lane)
    linear_particle.spawn(linear_layout, NOTE_PARTICLE_DURATION)
    circular_particle = get_note_circular_particle(kind, direction)
    circular_layout = layout_note_circular_particle(lane)
    circular_particle.spawn(circular_layout, NOTE_PARTICLE_DURATION)
    play_lane_particle(lane)


def get_note_sfx(kind: NoteKind, judgment: int) -> Effect:
    result = +Effect
    match kind:
        case NoteKind.FLICK | NoteKind.DIRECTIONAL_FLICK:
            match judgment:
                case Judgment.PERFECT:
                    result @= Effects.perfect_alt
                case Judgment.GREAT:
                    result @= Effects.great_alt
                case Judgment.GOOD:
                    result @= Effects.good_alt
        case _:
            match judgment:
                case Judgment.PERFECT:
                    result @= Effects.perfect
                case Judgment.GREAT:
                    result @= Effects.great
                case Judgment.GOOD:
                    result @= Effects.good
    return result


def play_note_sfx(kind: NoteKind, judgment: int):
    if not Options.sfx_enabled:
        return
    if kind == NoteKind.HOLD_ANCHOR:
        return
    sfx = get_note_sfx(kind, judgment)
    sfx.play()


def schedule_note_sfx(kind: NoteKind, judgment: int, target_time: float):
    if not Options.sfx_enabled:
        return
    if kind == NoteKind.HOLD_ANCHOR:
        return
    sfx = get_note_sfx(kind, judgment)
    sfx.schedule(target_time)


def draw_note_head(lane: float):
    """Draw an active hold note head at the judgment line."""
    layout = layout_note_body(lane, 0)
    Skin.hold_head_note.draw(layout, z=get_z(LAYER_NOTE_HEAD, lane=lane, y=0), a=1)


def update_hold_particle(handle: ParticleHandle, lane: float):
    """Update an active hold note particle effect handle to the current lane, spawning a new particle if needed."""
    layout = layout_note_circular_particle(lane)
    if handle.id == 0:
        handle @= Particles.hold_active_circular.spawn(layout, ACTIVE_HOLD_PARTICLE_DURATION, loop=True)
    else:
        handle.move(layout)


def destroy_particle(handle: ParticleHandle):
    """Destroy a particle effect if it exists."""
    if handle.id != 0:
        handle.destroy()
        handle.id = 0


def update_hold_sfx(handle: LoopedEffectHandle):
    """Start an active hold note looped sound effect if it's not already playing."""
    if not Options.sfx_enabled:
        return
    if handle.id == 0:
        handle @= Effects.hold.loop()


def stop_looped_sfx(handle: LoopedEffectHandle):
    """Stop a looped sound effect if it's playing."""
    if handle.id != 0:
        handle.stop()
        handle.id = 0


def schedule_hold_sfx(start_time: float, end_time: float):
    """Schedule an active hold note looped sound effect to play over an interval."""
    if not Options.sfx_enabled:
        return
    Effects.hold.schedule_loop(start_time).stop(end_time)


def init_note_life(archetype: type[PlayArchetype | WatchArchetype]):
    match archetype.key:
        case NoteKind.HOLD_TICK:
            archetype.life.update(
                perfect_increment=1,
                miss_increment=-20,
            )
        case _:
            archetype.life.update(
                perfect_increment=1,
                miss_increment=-100,
            )


def get_flick_speed_threshold(direction: int) -> float:
    """Return the speed threshold for a flick input to be counted."""
    if direction == 0:
        return FLICK_SPEED_THRESHOLD * Layout.lane_width
    else:
        return (
            DIRECTIONAL_FLICK_SPEED_THRESHOLD_BASE + DIRECTIONAL_FLICK_SPEED_THRESHOLD_INCREMENT * abs(direction)
        ) * Layout.lane_width



from enum import StrEnum


class StandardText(StrEnum):
    """Standard text constants."""

    CUSTOM_SERVER = "#CUSTOM_SERVER"
    """Custom Server"""

    COLLECTION = "#COLLECTION"
    """Collection"""

    SERVER = "#SERVER"
    """Server"""

    ADDRESS = "#ADDRESS"
    """Address"""

    EXPIRATION = "#EXPIRATION"
    """Expiration"""

    STORAGE = "#STORAGE"
    """Storage"""

    LOG = "#LOG"
    """Log"""

    INQUIRY = "#INQUIRY"
    """Inquiry"""

    BANNER = "#BANNER"
    """Banner"""

    POST = "#POST"
    """Post"""

    PLAYLIST = "#PLAYLIST"
    """Playlist"""

    LEVEL = "#LEVEL"
    """Level"""

    SKIN = "#SKIN"
    """Skin"""

    BACKGROUND = "#BACKGROUND"
    """Background"""

    EFFECT = "#EFFECT"
    """SFX"""

    PARTICLE = "#PARTICLE"
    """Particle"""

    ENGINE = "#ENGINE"
    """Engine"""

    REPLAY = "#REPLAY"
    """Replay"""

    USER = "#USER"
    """User"""

    ROOM = "#ROOM"
    """Room"""

    POST_THUMBNAIL = "#POST_THUMBNAIL"
    """Thumbnail"""

    PLAYLIST_THUMBNAIL = "#PLAYLIST_THUMBNAIL"
    """Thumbnail"""

    LEVEL_COVER = "#LEVEL_COVER"
    """Cover"""

    LEVEL_BGM = "#LEVEL_BGM"
    """BGM"""

    LEVEL_PREVIEW = "#LEVEL_PREVIEW"
    """Preview"""

    LEVEL_DATA = "#LEVEL_DATA"
    """Data"""

    SKIN_THUMBNAIL = "#SKIN_THUMBNAIL"
    """Thumbnail"""

    SKIN_DATA = "#SKIN_DATA"
    """Data"""

    SKIN_TEXTURE = "#SKIN_TEXTURE"
    """Texture"""

    BACKGROUND_THUMBNAIL = "#BACKGROUND_THUMBNAIL"
    """Thumbnail"""

    BACKGROUND_IMAGE = "#BACKGROUND_IMAGE"
    """Image"""

    BACKGROUND_DATA = "#BACKGROUND_DATA"
    """Data"""

    BACKGROUND_CONFIGURATION = "#BACKGROUND_CONFIGURATION"
    """Configuration"""

    EFFECT_THUMBNAIL = "#EFFECT_THUMBNAIL"
    """Thumbnail"""

    EFFECT_DATA = "#EFFECT_DATA"
    """Data"""

    EFFECT_AUDIO = "#EFFECT_AUDIO"
    """Audio"""

    PARTICLE_THUMBNAIL = "#PARTICLE_THUMBNAIL"
    """Thumbnail"""

    PARTICLE_DATA = "#PARTICLE_DATA"
    """Data"""

    PARTICLE_TEXTURE = "#PARTICLE_TEXTURE"
    """Texture"""

    ENGINE_THUMBNAIL = "#ENGINE_THUMBNAIL"
    """Thumbnail"""

    ENGINE_PLAYDATA = "#ENGINE_PLAYDATA"
    """Play Data"""

    ENGINE_WATCHDATA = "#ENGINE_WATCHDATA"
    """Watch Data"""

    ENGINE_PREVIEWDATA = "#ENGINE_PREVIEWDATA"
    """Preview Data"""

    ENGINE_TUTORIALDATA = "#ENGINE_TUTORIALDATA"
    """Tutorial Data"""

    ENGINE_ROM = "#ENGINE_ROM"
    """ROM"""

    ENGINE_CONFIGURATION = "#ENGINE_CONFIGURATION"
    """Configuration"""

    REPLAY_DATA = "#REPLAY_DATA"
    """Data"""

    REPLAY_CONFIGURATION = "#REPLAY_CONFIGURATION"
    """Configuration"""

    ROOM_COVER = "#ROOM_COVER"
    """Cover"""

    ROOM_BGM = "#ROOM_BGM"
    """BGM"""

    ROOM_PREVIEW = "#ROOM_PREVIEW"
    """Preview"""

    GRADE = "#GRADE"
    """Grade"""

    ARCADE_SCORE = "#ARCADE_SCORE"
    """Arcade Score"""

    ACCURACY_SCORE = "#ACCURACY_SCORE"
    """Accuracy Score"""

    COMBO = "#COMBO"
    """Combo"""

    PERFECT = "#PERFECT"
    """Perfect"""

    GREAT = "#GREAT"
    """Great"""

    GOOD = "#GOOD"
    """Good"""

    MISS = "#MISS"
    """Miss"""

    JUDGMENT = "#JUDGMENT"
    """Judgment"""

    ACCURACY = "#ACCURACY"
    """Accuracy"""

    FILTER = "#FILTER"
    """Filter"""

    SORT = "#SORT"
    """Sort"""

    KEYWORDS = "#KEYWORDS"
    """Keywords"""

    NAME = "#NAME"
    """Name"""

    RATING = "#RATING"
    """Rating"""

    RATING_MINIMUM = "#RATING_MINIMUM"
    """Minimum Rating"""

    RATING_MAXIMUM = "#RATING_MAXIMUM"
    """Maximum Rating"""

    TITLE = "#TITLE"
    """Title"""

    SUBTITLE = "#SUBTITLE"
    """Subtitle"""

    ARTISTS = "#ARTISTS"
    """Artists"""

    TIME = "#TIME"
    """Time"""

    AUTHOR = "#AUTHOR"
    """Author"""

    COAUTHOR = "#COAUTHOR"
    """Coauthor"""

    DESCRIPTION = "#DESCRIPTION"
    """Description"""

    GENRE = "#GENRE"
    """Genre"""

    TYPE = "#TYPE"
    """Type"""

    CATEGORY = "#CATEGORY"
    """Category"""

    STATUS = "#STATUS"
    """Status"""

    LANGUAGE = "#LANGUAGE"
    """Language"""

    DIFFICULTY = "#DIFFICULTY"
    """Difficulty"""

    VERSION = "#VERSION"
    """Version"""

    LENGTH = "#LENGTH"
    """Length"""

    LENGTH_MINIMUM = "#LENGTH_MINIMUM"
    """Minimum Length"""

    LENGTH_MAXIMUM = "#LENGTH_MAXIMUM"
    """Maximum Length"""

    ADDITIONAL_INFORMATION = "#ADDITIONAL_INFORMATION"
    """Additional Information"""

    TIMEZONE = "#TIMEZONE"
    """Timezone"""

    REGION = "#REGION"
    """Region"""

    TAG = "#TAG"
    """Tag"""

    INCLUDE_TAG = "#INCLUDE_TAG"
    """Include Tag"""

    EXCLUDE_TAG = "#EXCLUDE_TAG"
    """Exclude Tag"""

    CONTENT = "#CONTENT"
    """Content"""

    COMMENT = "#COMMENT"
    """Comment"""

    MESSAGE = "#MESSAGE"
    """Message"""

    NOTIFICATION = "#NOTIFICATION"
    """Notification"""

    ROLE = "#ROLE"
    """Role"""

    PERMISSION = "#PERMISSION"
    """Permission"""

    SPEED = "#SPEED"
    """Level Speed"""

    MIRROR = "#MIRROR"
    """Mirror Level"""

    RANDOM = "#RANDOM"
    """Random"""

    HIDDEN = "#HIDDEN"
    """Hidden"""

    JUDGMENT_STRICT = "#JUDGMENT_STRICT"
    """Strict Judgment"""

    JUDGMENT_LOOSE = "#JUDGMENT_LOOSE"
    """Loose Judgment"""

    EFFECT_AUTO = "#EFFECT_AUTO"
    """Auto SFX"""

    HAPTIC = "#HAPTIC"
    """Haptic"""

    STAGE = "#STAGE"
    """Stage"""

    STAGE_POSITION = "#STAGE_POSITION"
    """Stage Position"""

    STAGE_SIZE = "#STAGE_SIZE"
    """Stage Size"""

    STAGE_ROTATION = "#STAGE_ROTATION"
    """Stage Rotation"""

    STAGE_DIRECTION = "#STAGE_DIRECTION"
    """Stage Direction"""

    STAGE_ALPHA = "#STAGE_ALPHA"
    """Stage Transparency"""

    STAGE_ANIMATION = "#STAGE_ANIMATION"
    """Stage Animation"""

    STAGE_TILT = "#STAGE_TILT"
    """Stage Tilt"""

    STAGE_COVER_VERTICAL = "#STAGE_COVER_VERTICAL"
    """Vertical Stage Cover"""

    STAGE_COVER_HORIZONTAL = "#STAGE_COVER_HORIZONTAL"
    """Horizontal Stage Cover"""

    STAGE_COVER_ALPHA = "#STAGE_COVER_ALPHA"
    """Stage Cover Transparency"""

    STAGE_ASPECTRATIO_LOCK = "#STAGE_ASPECTRATIO_LOCK"
    """Lock Stage Aspect Ratio"""

    STAGE_EFFECT = "#STAGE_EFFECT"
    """Stage Effect"""

    STAGE_EFFECT_POSITION = "#STAGE_EFFECT_POSITION"
    """Stage Effect Position"""

    STAGE_EFFECT_SIZE = "#STAGE_EFFECT_SIZE"
    """Stage Effect Size"""

    STAGE_EFFECT_ROTATION = "#STAGE_EFFECT_ROTATION"
    """Stage Effect Rotation"""

    STAGE_EFFECT_DIRECTION = "#STAGE_EFFECT_DIRECTION"
    """Stage Effect Direction"""

    STAGE_EFFECT_ALPHA = "#STAGE_EFFECT_ALPHA"
    """Stage Effect Transparency"""

    LANE = "#LANE"
    """Lane"""

    LANE_POSITION = "#LANE_POSITION"
    """Lane Position"""

    LANE_SIZE = "#LANE_SIZE"
    """Lane Size"""

    LANE_ROTATION = "#LANE_ROTATION"
    """Lane Rotation"""

    LANE_DIRECTION = "#LANE_DIRECTION"
    """Lane Direction"""

    LANE_ALPHA = "#LANE_ALPHA"
    """Lane Transparency"""

    LANE_ANIMATION = "#LANE_ANIMATION"
    """Lane Animation"""

    LANE_EFFECT = "#LANE_EFFECT"
    """Lane Effect"""

    LANE_EFFECT_POSITION = "#LANE_EFFECT_POSITION"
    """Lane Effect Position"""

    LANE_EFFECT_SIZE = "#LANE_EFFECT_SIZE"
    """Lane Effect Size"""

    LANE_EFFECT_ROTATION = "#LANE_EFFECT_ROTATION"
    """Lane Effect Rotation"""

    LANE_EFFECT_DIRECTION = "#LANE_EFFECT_DIRECTION"
    """Lane Effect Direction"""

    LANE_EFFECT_ALPHA = "#LANE_EFFECT_ALPHA"
    """Lane Effect Transparency"""

    JUDGELINE = "#JUDGELINE"
    """Judgment Line"""

    JUDGELINE_POSITION = "#JUDGELINE_POSITION"
    """Judgment Line Position"""

    JUDGELINE_SIZE = "#JUDGELINE_SIZE"
    """Judgment Line Size"""

    JUDGELINE_ROTATION = "#JUDGELINE_ROTATION"
    """Judgment Line Rotation"""

    JUDGELINE_DIRECTION = "#JUDGELINE_DIRECTION"
    """Judgment Line Direction"""

    JUDGELINE_ALPHA = "#JUDGELINE_ALPHA"
    """Judgment Line Transparency"""

    JUDGELINE_ANIMATION = "#JUDGELINE_ANIMATION"
    """Judgment Line Animation"""

    JUDGELINE_EFFECT = "#JUDGELINE_EFFECT"
    """Judgment Line Effect"""

    JUDGELINE_EFFECT_POSITION = "#JUDGELINE_EFFECT_POSITION"
    """Judgment Line Effect Position"""

    JUDGELINE_EFFECT_SIZE = "#JUDGELINE_EFFECT_SIZE"
    """Judgment Line Effect Size"""

    JUDGELINE_EFFECT_ROTATION = "#JUDGELINE_EFFECT_ROTATION"
    """Judgment Line Effect Rotation"""

    JUDGELINE_EFFECT_DIRECTION = "#JUDGELINE_EFFECT_DIRECTION"
    """Judgment Line Effect Direction"""

    JUDGELINE_EFFECT_ALPHA = "#JUDGELINE_EFFECT_ALPHA"
    """Judgment Line Effect Transparency"""

    SLOT = "#SLOT"
    """Slot"""

    SLOT_POSITION = "#SLOT_POSITION"
    """Slot Position"""

    SLOT_SIZE = "#SLOT_SIZE"
    """Slot Size"""

    SLOT_ROTATION = "#SLOT_ROTATION"
    """Slot Rotation"""

    SLOT_DIRECTION = "#SLOT_DIRECTION"
    """Slot Direction"""

    SLOT_ALPHA = "#SLOT_ALPHA"
    """Slot Transparency"""

    SLOT_ANIMATION = "#SLOT_ANIMATION"
    """Slot Animation"""

    SLOT_EFFECT = "#SLOT_EFFECT"
    """Slot Effect"""

    SLOT_EFFECT_POSITION = "#SLOT_EFFECT_POSITION"
    """Slot Effect Position"""

    SLOT_EFFECT_SIZE = "#SLOT_EFFECT_SIZE"
    """Slot Effect Size"""

    SLOT_EFFECT_ROTATION = "#SLOT_EFFECT_ROTATION"
    """Slot Effect Rotation"""

    SLOT_EFFECT_DIRECTION = "#SLOT_EFFECT_DIRECTION"
    """Slot Effect Direction"""

    SLOT_EFFECT_ALPHA = "#SLOT_EFFECT_ALPHA"
    """Slot Effect Transparency"""

    NOTE = "#NOTE"
    """Note"""

    NOTE_SPEED = "#NOTE_SPEED"
    """Note Speed"""

    NOTE_SPEED_RANDOM = "#NOTE_SPEED_RANDOM"
    """Random Note Speed"""

    NOTE_POSITION = "#NOTE_POSITION"
    """Note Position"""

    NOTE_SIZE = "#NOTE_SIZE"
    """Note Size"""

    NOTE_ROTATION = "#NOTE_ROTATION"
    """Note Rotation"""

    NOTE_DIRECTION = "#NOTE_DIRECTION"
    """Note Direction"""

    NOTE_COLOR = "#NOTE_COLOR"
    """Note Color"""

    NOTE_ALPHA = "#NOTE_ALPHA"
    """Note Transparency"""

    NOTE_ANIMATION = "#NOTE_ANIMATION"
    """Note Animation"""

    NOTE_EFFECT = "#NOTE_EFFECT"
    """Note Effect"""

    NOTE_EFFECT_POSITION = "#NOTE_EFFECT_POSITION"
    """Note Effect Position"""

    NOTE_EFFECT_SIZE = "#NOTE_EFFECT_SIZE"
    """Note Effect Size"""

    NOTE_EFFECT_ROTATION = "#NOTE_EFFECT_ROTATION"
    """Note Effect Rotation"""

    NOTE_EFFECT_DIRECTION = "#NOTE_EFFECT_DIRECTION"
    """Note Effect Direction"""

    NOTE_EFFECT_COLOR = "#NOTE_EFFECT_COLOR"
    """Note Effect Color"""

    NOTE_EFFECT_ALPHA = "#NOTE_EFFECT_ALPHA"
    """Note Effect Transparency"""

    MARKER = "#MARKER"
    """Marker"""

    MARKER_POSITION = "#MARKER_POSITION"
    """Marker Position"""

    MARKER_SIZE = "#MARKER_SIZE"
    """Marker Size"""

    MARKER_ROTATION = "#MARKER_ROTATION"
    """Marker Rotation"""

    MARKER_DIRECTION = "#MARKER_DIRECTION"
    """Marker Direction"""

    MARKER_COLOR = "#MARKER_COLOR"
    """Marker Color"""

    MARKER_ALPHA = "#MARKER_ALPHA"
    """Marker Transparency"""

    MARKER_ANIMATION = "#MARKER_ANIMATION"
    """Marker Animation"""

    CONNECTOR = "#CONNECTOR"
    """Connector"""

    CONNECTOR_POSITION = "#CONNECTOR_POSITION"
    """Connector Position"""

    CONNECTOR_SIZE = "#CONNECTOR_SIZE"
    """Connector Size"""

    CONNECTOR_ROTATION = "#CONNECTOR_ROTATION"
    """Connector Rotation"""

    CONNECTOR_DIRECTION = "#CONNECTOR_DIRECTION"
    """Connector Direction"""

    CONNECTOR_COLOR = "#CONNECTOR_COLOR"
    """Connector Color"""

    CONNECTOR_ALPHA = "#CONNECTOR_ALPHA"
    """Connector Transparency"""

    CONNECTOR_ANIMATION = "#CONNECTOR_ANIMATION"
    """Connector Animation"""

    SIMLINE = "#SIMLINE"
    """Simultaneous Line"""

    SIMLINE_POSITION = "#SIMLINE_POSITION"
    """Simultaneous Line Position"""

    SIMLINE_SIZE = "#SIMLINE_SIZE"
    """Simultaneous Line Size"""

    SIMLINE_ROTATION = "#SIMLINE_ROTATION"
    """Simultaneous Line Rotation"""

    SIMLINE_DIRECTION = "#SIMLINE_DIRECTION"
    """Simultaneous Line Direction"""

    SIMLINE_COLOR = "#SIMLINE_COLOR"
    """Simultaneous Line Color"""

    SIMLINE_ALPHA = "#SIMLINE_ALPHA"
    """Simultaneous Line Transparency"""

    SIMLINE_ANIMATION = "#SIMLINE_ANIMATION"
    """Simultaneous Line Animation"""

    PREVIEW_SCALE_VERTICAL = "#PREVIEW_SCALE_VERTICAL"
    """Preview Vertical Scale"""

    PREVIEW_SCALE_HORIZONTAL = "#PREVIEW_SCALE_HORIZONTAL"
    """Preview Horizontal Scale"""

    PREVIEW_TIME = "#PREVIEW_TIME"
    """Preview Time"""

    PREVIEW_SCORE = "#PREVIEW_SCORE"
    """Preview Score"""

    PREVIEW_BPM = "#PREVIEW_BPM"
    """Preview BPM"""

    PREVIEW_TIMESCALE = "#PREVIEW_TIMESCALE"
    """Preview Time Scale"""

    PREVIEW_BEAT = "#PREVIEW_BEAT"
    """Preview Beat"""

    PREVIEW_MEASURE = "#PREVIEW_MEASURE"
    """Preview Measure"""

    PREVIEW_COMBO = "#PREVIEW_COMBO"
    """Preview Combo"""

    UI = "#UI"
    """UI"""

    UI_METRIC = "#UI_METRIC"
    """UI Metric"""

    UI_PRIMARY_METRIC = "#UI_PRIMARY_METRIC"
    """UI Primary Metric"""

    UI_SECONDARY_METRIC = "#UI_SECONDARY_METRIC"
    """UI Secondary Metric"""

    UI_JUDGMENT = "#UI_JUDGMENT"
    """UI Judgment"""

    UI_COMBO = "#UI_Combo"
    """UI Combo"""

    UI_MENU = "#UI_Menu"
    """UI Menu"""

    ON = "#ON"
    """ON"""

    OFF = "#OFF"
    """OFF"""

    NONE = "#NONE"
    """None"""

    ANY = "#ANY"
    """Any"""

    ALL = "#ALL"
    """All"""

    OTHERS = "#OTHERS"
    """Others"""

    SHORT = "#SHORT"
    """Short"""

    LONG = "#LONG"
    """Long"""

    HIGH = "#HIGH"
    """High"""

    MID = "#MID"
    """Mid"""

    LOW = "#LOW"
    """Low"""

    SMALL = "#SMALL"
    """Small"""

    MEDIUM = "#MEDIUM"
    """Medium"""

    LARGE = "#LARGE"
    """Large"""

    LEFT = "#LEFT"
    """Left"""

    RIGHT = "#RIGHT"
    """Right"""

    UP = "#UP"
    """Up"""

    DOWN = "#DOWN"
    """Down"""

    FRONT = "#FRONT"
    """Front"""

    BACK = "#BACK"
    """Back"""

    CENTER = "#CENTER"
    """Center"""

    TOP = "#TOP"
    """Top"""

    BOTTOM = "#BOTTOM"
    """Bottom"""

    TOP_LEFT = "#TOP_LEFT"
    """Top Left"""

    TOP_CENTER = "#TOP_CENTER"
    """Top Center"""

    TOP_RIGHT = "#TOP_RIGHT"
    """Top Right"""

    CENTER_LEFT = "#CENTER_LEFT"
    """Center Left"""

    CENTER_RIGHT = "#CENTER_RIGHT"
    """Center Right"""

    BOTTOM_LEFT = "#BOTTOM_LEFT"
    """Bottom Left"""

    BOTTOM_CENTER = "#BOTTOM_CENTER"
    """Bottom Center"""

    BOTTOM_RIGHT = "#BOTTOM_RIGHT"
    """Bottom Right"""

    CLOCKWISE = "#CLOCKWISE"
    """Clockwise"""

    COUNTERCLOCKWISE = "#COUNTERCLOCKWISE"
    """Counterclockwise"""

    FORWARD = "#FORWARD"
    """Forward"""

    BACKWARD = "#BACKWARD"
    """Backward"""

    DEFAULT = "#DEFAULT"
    """Default"""

    NEUTRAL = "#NEUTRAL"
    """Neutral"""

    RED = "#RED"
    """Red"""

    GREEN = "#GREEN"
    """Green"""

    BLUE = "#BLUE"
    """Blue"""

    YELLOW = "#YELLOW"
    """Yellow"""

    PURPLE = "#PURPLE"
    """Purple"""

    CYAN = "#CYAN"
    """Cyan"""

    SIMPLE = "#SIMPLE"
    """Simple"""

    EASY = "#EASY"
    """Easy"""

    NORMAL = "#NORMAL"
    """Normal"""

    HARD = "#HARD"
    """Hard"""

    EXPERT = "#EXPERT"
    """Expert"""

    MASTER = "#MASTER"
    """Master"""

    PRO = "#PRO"
    """Pro"""

    TECHNICAL = "#TECHNICAL"
    """Technical"""

    SPECIAL = "#SPECIAL"
    """Special"""

    APPEND = "#APPEND"
    """Append"""

    POST_PLACEHOLDER = "#POST_PLACEHOLDER"
    """Enter post..."""

    PLAYLIST_PLACEHOLDER = "#PLAYLIST_PLACEHOLDER"
    """Enter playlist..."""

    LEVEL_PLACEHOLDER = "#LEVEL_PLACEHOLDER"
    """Enter level..."""

    SKIN_PLACEHOLDER = "#SKIN_PLACEHOLDER"
    """Enter skin..."""

    BACKGROUND_PLACEHOLDER = "#BACKGROUND_PLACEHOLDER"
    """Enter background..."""

    EFFECT_PLACEHOLDER = "#EFFECT_PLACEHOLDER"
    """Enter sFX..."""

    PARTICLE_PLACEHOLDER = "#PARTICLE_PLACEHOLDER"
    """Enter particle..."""

    ENGINE_PLACEHOLDER = "#ENGINE_PLACEHOLDER"
    """Enter engine..."""

    REPLAY_PLACEHOLDER = "#REPLAY_PLACEHOLDER"
    """Enter replay..."""

    USER_PLACEHOLDER = "#USER_PLACEHOLDER"
    """Enter user..."""

    ROOM_PLACEHOLDER = "#ROOM_PLACEHOLDER"
    """Enter room..."""

    KEYWORDS_PLACEHOLDER = "#KEYWORDS_PLACEHOLDER"
    """Enter keywords..."""

    NAME_PLACEHOLDER = "#NAME_PLACEHOLDER"
    """Enter name..."""

    RATING_PLACEHOLDER = "#RATING_PLACEHOLDER"
    """Enter rating..."""

    RATING_MINIMUM_PLACEHOLDER = "#RATING_MINIMUM_PLACEHOLDER"
    """Enter minimum rating..."""

    RATING_MAXIMUM_PLACEHOLDER = "#RATING_MAXIMUM_PLACEHOLDER"
    """Enter maximum rating..."""

    TITLE_PLACEHOLDER = "#TITLE_PLACEHOLDER"
    """Enter title..."""

    SUBTITLE_PLACEHOLDER = "#SUBTITLE_PLACEHOLDER"
    """Enter subtitle..."""

    ARTISTS_PLACEHOLDER = "#ARTISTS_PLACEHOLDER"
    """Enter artists..."""

    TIME_PLACEHOLDER = "#TIME_PLACEHOLDER"
    """Enter time..."""

    AUTHOR_PLACEHOLDER = "#AUTHOR_PLACEHOLDER"
    """Enter author..."""

    DESCRIPTION_PLACEHOLDER = "#DESCRIPTION_PLACEHOLDER"
    """Enter description..."""

    GENRE_PLACEHOLDER = "#GENRE_PLACEHOLDER"
    """Enter genre..."""

    TYPE_PLACEHOLDER = "#TYPE_PLACEHOLDER"
    """Enter type..."""

    CATEGORY_PLACEHOLDER = "#CATEGORY_PLACEHOLDER"
    """Enter category..."""

    LANGUAGE_PLACEHOLDER = "#LANGUAGE_PLACEHOLDER"
    """Enter language..."""

    DIFFICULTY_PLACEHOLDER = "#DIFFICULTY_PLACEHOLDER"
    """Enter difficulty..."""

    LENGTH_PLACEHOLDER = "#LENGTH_PLACEHOLDER"
    """Enter length..."""

    LENGTH_MINIMUM_PLACEHOLDER = "#LENGTH_MINIMUM_PLACEHOLDER"
    """Enter minimum length..."""

    LENGTH_MAXIMUM_PLACEHOLDER = "#LENGTH_MAXIMUM_PLACEHOLDER"
    """Enter maximum length..."""

    ADDITIONAL_INFORMATION_PLACEHOLDER = "#ADDITIONAL_INFORMATION_PLACEHOLDER"
    """Enter additional information..."""

    TIMEZONE_PLACEHOLDER = "#TIMEZONE_PLACEHOLDER"
    """Enter timezone..."""

    REGION_PLACEHOLDER = "#REGION_PLACEHOLDER"
    """Enter region..."""

    CONTENT_PLACEHOLDER = "#CONTENT_PLACEHOLDER"
    """Enter content..."""

    COMMENT_PLACEHOLDER = "#COMMENT_PLACEHOLDER"
    """Enter comment..."""

    REVIEW_PLACEHOLDER = "#REVIEW_PLACEHOLDER"
    """Enter review..."""

    REPLY_PLACEHOLDER = "#REPLY_PLACEHOLDER"
    """Enter reply..."""

    MESSAGE_PLACEHOLDER = "#MESSAGE_PLACEHOLDER"
    """Enter message..."""

    ROLE_PLACEHOLDER = "#ROLE_PLACEHOLDER"
    """Enter role..."""

    PERMISSION_PLACEHOLDER = "#PERMISSION_PLACEHOLDER"
    """Enter permission..."""

    PERCENTAGE_UNIT = "#PERCENTAGE_UNIT"
    """{0}%"""

    YEAR_UNIT = "#YEAR_UNIT"
    """{0} yr"""

    MONTH_UNIT = "#MONTH_UNIT"
    """{0} mo"""

    DAY_UNIT = "#DAY_UNIT"
    """{0} d"""

    HOUR_UNIT = "#HOUR_UNIT"
    """{0} h"""

    MINUTE_UNIT = "#MINUTE_UNIT"
    """{0} m"""

    SECOND_UNIT = "#SECOND_UNIT"
    """{0} s"""

    MILLISECOND_UNIT = "#MILLISECOND_UNIT"
    """{0} ms"""

    YEAR_PAST = "#YEAR_PAST"
    """{0} yr ago"""

    MONTH_PAST = "#MONTH_PAST"
    """{0} mo ago"""

    DAY_PAST = "#DAY_PAST"
    """{0} d ago"""

    HOUR_PAST = "#HOUR_PAST"
    """{0} h ago"""

    MINUTE_PAST = "#MINUTE_PAST"
    """{0} m ago"""

    SECOND_PAST = "#SECOND_PAST"
    """{0} s ago"""

    MILLISECOND_PAST = "#MILLISECOND_PAST"
    """{0} ms ago"""

    YEAR_FUTURE = "#YEAR_FUTURE"
    """In {0} yr"""

    MONTH_FUTURE = "#MONTH_FUTURE"
    """In {0} mo"""

    DAY_FUTURE = "#DAY_FUTURE"
    """In {0} d"""

    HOUR_FUTURE = "#HOUR_FUTURE"
    """In {0} h"""

    MINUTE_FUTURE = "#MINUTE_FUTURE"
    """In {0} m"""

    SECOND_FUTURE = "#SECOND_FUTURE"
    """In {0} s"""

    MILLISECOND_FUTURE = "#MILLISECOND_FUTURE"
    """In {0} ms"""

    TAP = "#TAP"
    """Tap"""

    TAP_HOLD = "#TAP_HOLD"
    """Tap and Hold"""

    TAP_RELEASE = "#TAP_RELEASE"
    """Tap and Release"""

    TAP_FLICK = "#TAP_FLICK"
    """Tap and Flick"""

    TAP_SLIDE = "#TAP_SLIDE"
    """Tap and Slide"""

    HOLD = "#HOLD"
    """Hold"""

    HOLD_SLIDE = "#HOLD_SLIDE"
    """Hold and Slide"""

    HOLD_FOLLOW = "#HOLD_FOLLOW"
    """Hold and Follow"""

    RELEASE = "#RELEASE"
    """Release"""

    FLICK = "#FLICK"
    """Flick"""

    SLIDE = "#SLIDE"
    """Slide"""

    SLIDE_FLICK = "#SLIDE_FLICK"
    """Slide and Flick"""

    AVOID = "#AVOID"
    """Avoid"""

    JIGGLE = "#JIGGLE"
    """Jiggle"""

    NEWEST = "#NEWEST"
    """Newest"""

    OLDEST = "#OLDEST"
    """Oldest"""

    RECOMMENDED = "#RECOMMENDED"
    """Recommended"""

    POPULAR = "#POPULAR"
    """Popular"""

    FEATURED = "#FEATURED"
    """Featured"""

    COMPETITIVE = "#COMPETITIVE"
    """Competitive"""

    TOURNAMENT = "#TOURNAMENT"
    """Tournament"""

    HOLIDAY = "#HOLIDAY"
    """Holiday"""

    LIMITED = "#LIMITED"
    """Limited"""

    ANNOUNCEMENT = "#ANNOUNCEMENT"
    """Announcement"""

    INFORMATION = "#INFORMATION"
    """Information"""

    HELP = "#HELP"
    """Help"""

    MAINTENANCE = "#MAINTENANCE"
    """Maintenance"""

    EVENT = "#EVENT"
    """Event"""

    UPDATE = "#UPDATE"
    """Update"""

    SEARCH = "#SEARCH"
    """Search"""

    ADVANCED = "#ADVANCED"
    """Advanced"""

    RELATED = "#RELATED"
    """Related"""

    SAME_AUTHOR = "#SAME_AUTHOR"
    """Same Author"""

    SAME_ARTISTS = "#SAME_ARTISTS"
    """Same Artists"""

    SAME_RATING = "#SAME_RATING"
    """Same Rating"""

    SAME_CATEGORY = "#SAME_CATEGORY"
    """Same Category"""

    SAME_DIFFICULTY = "#SAME_DIFFICULTY"
    """Same Difficulty"""

    SAME_GENRE = "#SAME_GENRE"
    """Same Genre"""

    SAME_VERSION = "#SAME_VERSION"
    """Same Version"""

    OTHER_AUTHORS = "#OTHER_AUTHORS"
    """Other Authors"""

    OTHER_ARTISTS = "#OTHER_ARTISTS"
    """Other Artists"""

    OTHER_RATINGS = "#OTHER_RATINGS"
    """Other Ratings"""

    OTHER_CATEGORIES = "#OTHER_CATEGORIES"
    """Other Categories"""

    OTHER_DIFFICULTIES = "#OTHER_DIFFICULTIES"
    """Other Difficulties"""

    OTHER_GENRES = "#OTHER_GENRES"
    """Other Genres"""

    OTHER_VERSIONS = "#OTHER_VERSIONS"
    """Other Versions"""

    DRAFT = "#DRAFT"
    """Draft"""

    PUBLIC = "#PUBLIC"
    """Public"""

    PRIVATE = "#PRIVATE"
    """Private"""

    POP = "#POP"
    """Pop"""

    ROCK = "#ROCK"
    """Rock"""

    HIPHOP = "#HIPHOP"
    """Hip Hop"""

    COUNTRY = "#COUNTRY"
    """Country"""

    ELECTRONIC = "#ELECTRONIC"
    """Electronic"""

    METAL = "#METAL"
    """Metal"""

    CLASSICAL = "#CLASSICAL"
    """Classical"""

    FOLK = "#FOLK"
    """Folk"""

    INDIE = "#INDIE"
    """Indie"""

    ANIME = "#ANIME"
    """Anime"""

    VOCALOID = "#VOCALOID"
    """Vocaloid"""

    REMIX = "#REMIX"
    """Remix"""

    INSTRUMENTAL = "#INSTRUMENTAL"
    """Instrumental"""

    SHORT_VERSION = "#SHORT_VERSION"
    """Short Version"""

    LONG_VERSION = "#LONG_VERSION"
    """Long Version"""

    CUT_VERSION = "#CUT_VERSION"
    """Cut Version"""

    FULL_VERSION = "#FULL_VERSION"
    """Full Version"""

    EXTENDED_VERSION = "#EXTENDED_VERSION"
    """Extended Version"""

    LIVE_VERSION = "#LIVE_VERSION"
    """Live Version"""

    EXPLICIT = "#EXPLICIT"
    """Explicit"""

    MULTI_FINGER = "#MULTI_FINGER"
    """Multi Finger"""

    FULL_HAND = "#FULL_HAND"
    """Full Hand"""

    CROSS_HAND = "#CROSS_HAND"
    """Cross Hand"""

    GIMMICK = "#GIMMICK"
    """Gimmick"""

    COLLABORATION = "#COLLABORATION"
    """Collaboration"""

    COLLABORATOR = "#COLLABORATOR"
    """Collaborator"""

    REPORT = "#REPORT"
    """Report"""

    REASON = "#REASON"
    """Reason"""

    ILLEGAL_ACTIVITIES = "#ILLEGAL_ACTIVITIES"
    """Illegal Activities"""

    CHEATING = "#CHEATING"
    """Cheating"""

    AFK = "#AFK"
    """AFK"""

    SPAMMING = "#SPAMMING"
    """Spamming"""

    VERBAL_ABUSE = "#VERBAL_ABUSE"
    """Verbal Abuse"""

    INAPPROPRIATE_LANGUAGE = "#INAPPROPRIATE_LANGUAGE"
    """Inappropriate Language"""

    NEGATIVE_ATTITUDE = "#NEGATIVE_ATTITUDE"
    """Negative Attitude"""

    DNF = "#DNF"
    """DNF"""

    SUGGESTIONS = "#SUGGESTIONS"
    """Suggestions"""

    SUGGESTIONS_PER_PLAYER = "#SUGGESTIONS_PER_PLAYER"
    """Suggestions per Player"""

    MATCH_SCORING = "#MATCH_SCORING"
    """Match Scoring"""

    MATCH_TIEBREAKER = "#MATCH_TIEBREAKER"
    """Match Tiebreaker"""

    MATCH_COUNT = "#MATCH_COUNT"
    """Match Count"""

    MATCH_LIMIT = "#MATCH_LIMIT"
    """Match Limit"""

    ROUND_SCORING = "#ROUND_SCORING"
    """Round Scoring"""

    ROUND_TIEBREAKER = "#ROUND_TIEBREAKER"
    """Round Tiebreaker"""

    ROUND_COUNT = "#ROUND_COUNT"
    """Round Count"""

    ROUND_LIMIT = "#ROUND_LIMIT"
    """Round Limit"""

    TEAM_SCORING = "#TEAM_SCORING"
    """Team Scoring"""

    TEAM_TIEBREAKER = "#TEAM_TIEBREAKER"
    """Team Tiebreaker"""

    TEAM_COUNT = "#TEAM_COUNT"
    """Team Count"""

    TEAM_LIMIT = "#TEAM_LIMIT"
    """Team Limit"""

    QUALIFIED = "#QUALIFIED"
    """Qualified"""

    DISQUALIFIED = "#DISQUALIFIED"
    """Disqualified"""

    RANKING = "#RANKING"
    """Ranking"""

    SCORE = "#SCORE"
    """Score"""

    OWNER = "#OWNER"
    """Owner"""

    ADMIN = "#ADMIN"
    """Admin"""

    MODERATOR = "#MODERATOR"
    """Moderator"""

    REVIEWER = "#REVIEWER"
    """Reviewer"""

    BANNED = "#BANNED"
    """Banned"""

    PLAYER = "#PLAYER"
    """Player"""

    SPECTATOR = "#SPECTATOR"
    """Spectator"""

    REFEREE = "#REFEREE"
    """Referee"""

    ELIMINATED = "#ELIMINATED"
    """Eliminated"""

    FINALIST = "#FINALIST"
    """Finalist"""

    FINISHED = "#FINISHED"
    """Finished"""

    WINNER = "#WINNER"
    """Winner"""

    GOLD_MEDAL = "#GOLD_MEDAL"
    """Gold Medal"""

    SILVER_MEDAL = "#SILVER_MEDAL"
    """Silver Medal"""

    BRONZE_MEDAL = "#BRONZE_MEDAL"
    """Bronze Medal"""

    TEAM1 = "#TEAM_1"
    """Team 1"""

    TEAM2 = "#TEAM_2"
    """Team 2"""

    TEAM3 = "#TEAM_3"
    """Team 3"""

    TEAM4 = "#TEAM_4"
    """Team 4"""

    TEAM5 = "#TEAM_5"
    """Team 5"""

    TEAM6 = "#TEAM_6"
    """Team 6"""

    TEAM7 = "#TEAM_7"
    """Team 7"""

    TEAM8 = "#TEAM_8"
    """Team 8"""

    TEAM_RED = "#TEAM_RED"
    """Team Red"""

    TEAM_GREEN = "#TEAM_GREEN"
    """Team Green"""

    TEAM_BLUE = "#TEAM_BLUE"
    """Team Blue"""

    TEAM_YELLOW = "#TEAM_YELLOW"
    """Team Yellow"""

    TEAM_PURPLE = "#TEAM_PURPLE"
    """Team Purple"""

    TEAM_CYAN = "#TEAM_CYAN"
    """Team Cyan"""

    TEAM_WHITE = "#TEAM_WHITE"
    """Team White"""

    TEAM_BLACK = "#TEAM_BLACK"
    """Team Black"""

    ADD = "#ADD"
    """Add"""

    ADDED = "#ADDED"
    """Added"""

    CREATE = "#CREATE"
    """Create"""

    CREATED = "#CREATED"
    """Created"""

    REPLY = "#REPLY"
    """Reply"""

    REPLIED = "#REPLIED"
    """Replied"""

    REVIEW = "#REVIEW"
    """Review"""

    REVIEWING = "#REVIEWING"
    """Reviewing"""

    REVIEWED = "#REVIEWED"
    """Reviewed"""

    VERIFY = "#VERIFY"
    """Verify"""

    VERIFYING = "#VERIFYING"
    """Verifying"""

    VERIFIED = "#VERIFIED"
    """Verified"""

    UPLOAD = "#UPLOAD"
    """Upload"""

    UPLOADING = "#UPLOADING"
    """Uploading"""

    UPLOADED = "#UPLOADED"
    """Uploaded"""

    SUBMIT = "#SUBMIT"
    """Submit"""

    SUBMITTING = "#SUBMITTING"
    """Submitting"""

    SUBMITTED = "#SUBMITTED"
    """Submitted"""

    EDIT = "#EDIT"
    """Edit"""

    EDITING = "#EDITING"
    """Editing"""

    EDITED = "#EDITED"
    """Edited"""

    LIKE = "#LIKE"
    """Like"""

    LIKED = "#LIKED"
    """Liked"""

    DISLIKE = "#DISLIKE"
    """Dislike"""

    DISLIKED = "#DISLIKED"
    """Disliked"""

    BOOKMARK = "#BOOKMARK"
    """Bookmark"""

    BOOKMARKED = "#BOOKMARKED"
    """Bookmarked"""

    DELETE = "#DELETE"
    """Delete"""

    DELETING = "#DELETING"
    """Deleting"""

    DELETED = "#DELETED"
    """Deleted"""

    REMOVE = "#REMOVE"
    """Remove"""

    REMOVING = "#REMOVING"
    """Removing"""

    REMOVED = "#REMOVED"
    """Removed"""

    RESTORE = "#RESTORE"
    """Restore"""

    RESTORING = "#RESTORING"
    """Restoring"""

    RESTORED = "#RESTORED"
    """Restored"""

    CONFIRM = "#CONFIRM"
    """Confirm"""

    CONFIRMED = "#CONFIRMED"
    """Confirmed"""

    CANCEL = "#CANCEL"
    """Cancel"""

    CANCELED = "#CANCELED"
    """Canceled"""

    INCREASE = "#INCREASE"
    """Increase"""

    DECREASE = "#DECREASE"
    """Decrease"""

    UPVOTE = "#UPVOTE"
    """Upvote"""

    UPVOTED = "#UPVOTED"
    """Upvoted"""

    DOWNVOTE = "#DOWNVOTE"
    """Downvote"""

    DOWNVOTED = "#DOWNVOTED"
    """Downvoted"""

    AGREE = "#AGREE"
    """Agree"""

    AGREED = "#AGREED"
    """Agreed"""

    DISAGREE = "#DISAGREE"
    """Disagree"""

    DISAGREED = "#DISAGREED"
    """Disagreed"""

    LOCK = "#LOCK"
    """Lock"""

    LOCKED = "#LOCKED"
    """Locked"""

    UNLOCK = "#UNLOCK"
    """Unlock"""

    UNLOCKED = "#UNLOCKED"
    """Unlocked"""

    PIN = "#PIN"
    """Pin"""

    PINNED = "#PINNED"
    """Pinned"""

    UNPIN = "#UNPIN"
    """Unpin"""

    UNPINNED = "#UNPINNED"
    """Unpinned"""

    FOLLOW = "#FOLLOW"
    """Follow"""

    FOLLOWING = "#FOLLOWING"
    """Following"""

    FOLLOWED = "#FOLLOWED"
    """Followed"""

    UNFOLLOW = "#UNFOLLOW"
    """Unfollow"""

    SUBSCRIBE = "#SUBSCRIBE"
    """Subscribe"""

    SUBSCRIBING = "#SUBSCRIBING"
    """Subscribing"""

    SUBSCRIBED = "#SUBSCRIBED"
    """Subscribed"""

    UNSUBSCRIBE = "#UNSUBSCRIBE"
    """Unsubscribe"""

    PUBLISH = "#PUBLISH"
    """Publish"""

    PUBLISHING = "#PUBLISHING"
    """Publishing"""

    PUBLISHED = "#PUBLISHED"
    """Published"""

    UNPUBLISH = "#UNPUBLISH"
    """Unpublish"""

    SHOW = "#SHOW"
    """Show"""

    HIDE = "#HIDE"
    """Hide"""

    ALLOW = "#ALLOW"
    """Allow"""

    ALLOWED = "#ALLOWED"
    """Allowed"""

    DISALLOW = "#DISALLOW"
    """Disallow"""

    DISALLOWED = "#DISALLOWED"
    """Disallowed"""

    APPROVE = "#APPROVE"
    """Approve"""

    APPROVED = "#APPROVED"
    """Approved"""

    DENY = "#DENY"
    """Deny"""

    DENIED = "#DENIED"
    """Denied"""

    ACCEPT = "#ACCEPT"
    """Accept"""

    ACCEPTED = "#ACCEPTED"
    """Accepted"""

    REJECT = "#REJECT"
    """Reject"""

    REJECTED = "#REJECTED"
    """Rejected"""

    STAR = "#STAR"
    """Star"""

    STARRED = "#STARRED"
    """Starred"""



# ruff: noqa: SIM113, PLC2701, C405

import pytest

from sonolus.script.array import Array
from sonolus.script.internal.context import ctx
from sonolus.script.internal.impl import validate_value
from sonolus.script.internal.math_impls import _floor
from sonolus.script.internal.meta_fn import meta_fn
from sonolus.script.internal.random import _random
from sonolus.script.internal.set_impl import SetImpl
from sonolus.script.internal.tuple_impl import TupleImpl
from sonolus.script.num import _is_num
from tests.script.conftest import run_and_validate


@meta_fn
def bb(*x):
    if len(x) == 1:
        x = x[0]
    if not ctx():
        return x
    x = validate_value(x)
    if _is_num(x):
        return x + _floor(_random())
    elif isinstance(x, TupleImpl):
        return TupleImpl(tuple(bb(e) for e in x.value))
    else:
        return x


# __contains__


def test_contains_present_small_size_string_key():
    def fn():
        s = {"a", "b"}
        return Array(
            "a" in s,
            "b" in s,
        )

    assert run_and_validate(fn) == Array(True, True)


def test_contains_present_small_size_numeric_key():
    def fn():
        s = {1, 2}
        return Array(1 in s, 2 in s)

    assert run_and_validate(fn) == Array(True, True)


def test_contains_present_small_size_tuple_key():
    def fn():
        s = {(1, 1), (2, 2)}
        return Array((1, 1) in s, (2, 2) in s)

    assert run_and_validate(fn) == Array(True, True)


def test_contains_present_small_size_mixed_key():
    def fn():
        s = {"a", 2, (3, 3)}
        return Array("a" in s, 2 in s, (3, 3) in s)

    assert run_and_validate(fn) == Array(True, True, True)


def test_contains_present_large_size_numeric_key():
    s = {20, 3, 15, 7, 25, 1, 18, 9, 22, 5, 12, 16, 8, 24, 2, 19, 11, 14, 6, 23, 4, 17, 13, 21, 10}

    def fn():
        return Array(1 in s, 13 in s, 25 in s)

    assert run_and_validate(fn) == Array(True, True, True)


def test_contains_present_large_size_string_key():
    s = set("azbycxdwevfugthsirjqkplomn")

    def fn():
        return Array("a" in s, "m" in s, "z" in s)

    assert run_and_validate(fn) == Array(True, True, True)


def test_contains_present_large_size_tuple_key():
    s = {(k, k) for k in [20, 3, 15, 7, 25, 1, 18, 9, 22, 5, 12, 16, 8, 24, 2, 19, 11, 14, 6, 23, 4, 17, 13, 21, 10]}

    def fn():
        return Array((1, 1) in s, (13, 13) in s, (25, 25) in s)

    assert run_and_validate(fn) == Array(True, True, True)


def test_contains_present_large_size_mixed_key():
    s = {"a", "b", "c", "d", "e", "f", "g", 1, 2, 3, 4, 5, 6, 7, (1, 1), (2, 2), (3, 3), (4, 4), (5, 5), (6, 6), (7, 7)}

    def fn():
        return Array("a" in s, 4 in s, (4, 4) in s)

    assert run_and_validate(fn) == Array(True, True, True)


def test_contains_absent_empty():
    s = set()

    def fn():
        return "a" in s

    assert not run_and_validate(fn)


def test_contains_absent_small_size_string_key():
    def fn():
        s = {"a", "b"}
        return Array("c" in s, "d" in s)

    assert run_and_validate(fn) == Array(False, False)


def test_contains_absent_small_size_numeric_key():
    def fn():
        s = {1, 2}
        return Array(3 in s, 4 in s)

    assert run_and_validate(fn) == Array(False, False)


def test_contains_absent_large_size_numeric_key():
    s = {20, 3, 15, 7, 25, 1, 18, 9, 22, 5, 12, 16, 8, 24, 2, 19, 11, 14, 6, 23, 4, 17, 13, 21, 10}

    def fn():
        return Array(0 in s, 26 in s, -1 in s)

    assert run_and_validate(fn) == Array(False, False, False)


def test_contains_with_runtime_key_small_size_numeric():
    s = {1, 2, 3}

    def fn():
        return Array(bb(1) in s, bb(2) in s, bb(4) in s)

    assert run_and_validate(fn) == Array(True, True, False)


def test_contains_with_runtime_key_large_size_numeric():
    s = {20, 3, 15, 7, 25, 1, 18, 9, 22, 5, 12, 16, 8, 24, 2, 19, 11, 14, 6, 23, 4, 17, 13, 21, 10}

    def fn():
        return Array(bb(1) in s, bb(13) in s, bb(26) in s)

    assert run_and_validate(fn) == Array(True, True, False)


# __len__


def test_len_empty():
    s = set()

    def fn():
        return len(s)

    assert run_and_validate(fn) == 0


def test_len_small():
    def fn():
        s = {"a", "b", "c"}
        return len(s)

    assert run_and_validate(fn) == 3


def test_len_large():
    s = set(range(25))

    def fn():
        return len(s)

    assert run_and_validate(fn) == 25


# __iter__


def test_iter_small_size_numeric_key():
    s = {1, 2, 3}

    def fn():
        results = +Array[int, 3]
        i = 0
        for k in s:
            results[i] = k
            i += 1
        return results

    assert sorted(run_and_validate(fn)) == sorted(s)


def test_iter_small_size_string_key():
    s = {"a", "b", "c"}

    def fn():
        results = +Array[int, 3]
        i = 0
        for _k in s:
            results[i] = "a" in s
            i += 1
        return results

    assert run_and_validate(fn) == Array(True, True, True)


def test_iter_large_size_numeric_key():
    s = {20, 3, 15, 7, 25, 1, 18, 9, 22, 5, 12, 16, 8, 24, 2, 19, 11, 14, 6, 23, 4, 17, 13, 21, 10}

    def fn():
        results = +Array[int, 25]
        i = 0
        for k in s:
            results[i] = k
            i += 1
        return results

    assert sorted(run_and_validate(fn)) == sorted(s)


def test_iter_large_size_mixed_key():
    s = {"a", "b", "c", "d", "e", "f", "g", 1, 2, 3, 4, 5, 6, 7, (1, 1), (2, 2), (3, 3), (4, 4), (5, 5), (6, 6), (7, 7)}

    def fn():
        count = 0
        for _ in s:
            count += 1
        return count

    assert run_and_validate(fn) == len(s)


def test_iter_empty():
    s = set()

    def fn():
        count = 0
        for _ in s:
            count += 1
        return count

    assert run_and_validate(fn) == 0


# __or__


def test_or_disjoint():
    s1 = {1, 2, 3}
    s2 = {4, 5, 6}

    def fn():
        s3 = s1 | s2
        results = +Array[int, 6]
        i = 0
        for k in s3:
            results[i] = k
            i += 1
        return results

    assert sorted(run_and_validate(fn)) == sorted(s1 | s2)


def test_or_overlapping():
    s1 = {1, 2, 3}
    s2 = {2, 3, 4}

    def fn():
        s3 = s1 | s2
        results = +Array[int, 4]
        i = 0
        for k in s3:
            results[i] = k
            i += 1
        return results

    assert sorted(run_and_validate(fn)) == sorted(s1 | s2)


def test_or_with_empty():
    s1 = {1, 2, 3}
    s2: frozenset = frozenset()

    def fn():
        s3 = s1 | s2
        results = +Array[int, 3]
        i = 0
        for k in s3:
            results[i] = k
            i += 1
        return results

    assert sorted(run_and_validate(fn)) == sorted(s1)


def test_or_contains_check():
    s1 = {1, 2, 3}
    s2 = {4, 5, 6}

    def fn():
        s3 = s1 | s2
        return Array(1 in s3, 4 in s3, 7 in s3)

    assert run_and_validate(fn) == Array(True, True, False)


def test_or_large():
    s1 = {20, 3, 15, 7, 25, 1, 18, 9, 22, 5, 12}
    s2 = {16, 8, 24, 2, 19, 11, 14, 6, 23, 4, 17, 13, 21, 10}

    def fn():
        s3 = s1 | s2
        total = 0
        for k in s3:
            total += k
        return total

    assert run_and_validate(fn) == sum(s1 | s2)


# __eq__


def test_eq_raises():
    # Use SetImpl instances directly so Python mode also raises TypeError
    s1 = SetImpl.from_set({1, 2})
    s2 = SetImpl.from_set({1, 2})

    def fn():
        return s1 == s2

    with pytest.raises(TypeError):
        run_and_validate(fn)


def test_set_empty():
    def fn():
        s = set()
        return len(s)

    assert run_and_validate(fn) == 0


def test_set_from_tuple():
    def fn():
        s = set((1, 2, 3))
        return Array(1 in s, 2 in s, 4 in s)

    assert run_and_validate(fn) == Array(True, True, False)


def test_set_from_string_tuple():
    def fn():
        s = set(("a", "b", "c"))
        return Array("a" in s, "b" in s, "d" in s)

    assert run_and_validate(fn) == Array(True, True, False)


def test_set_with_type_params():
    def fn():
        s = set[int]((1, 2, 3))
        return Array(1 in s, 2 in s, 4 in s)

    assert run_and_validate(fn) == Array(True, True, False)


def test_set_copy():
    def fn():
        s1 = {1, 2, 3}
        s2 = set(s1)
        return Array(1 in s2, 2 in s2, 4 in s2)

    assert run_and_validate(fn) == Array(True, True, False)


def test_isinstance_set():
    s = {"a", "b"}

    def fn():
        return isinstance(s, set)

    assert run_and_validate(fn)


def test_isinstance_set_not_dict():
    d = {"a": 1}

    def fn():
        return isinstance(d, set)

    assert not run_and_validate(fn)
